from __future__ import annotations

import scipy as sp


def lowpass(data, fc, sample_rate, method="bessel", order=1, axis=-1):
    if method == "bessel":
        return bessel_lowpass(
            data=data,
            fc=fc,
            sample_rate=sample_rate,
            order=order,
            axis=axis,
        )  # noqa
    else:
        raise ValueError(f"Invalid method '{method}'.")


def highpass(data, fc, sample_rate, method="bessel", order=1, axis=-1):
    if method == "bessel":
        return bessel_highpass(
            data=data,
            fc=fc,
            sample_rate=sample_rate,
            order=order,
            axis=axis,
        )  # noqa
    else:
        raise ValueError(f"Invalid method '{method}'.")


def bandpass(data, f_lower, f_upper, sample_rate, method="bessel", order=1, axis=-1):
    kwargs = {
        "sample_rate": sample_rate,
        "order": order,
        "axis": axis,
        "method": method,
    }  # noqa
    if method == "bessel":
        return bessel_highpass(
            bessel_lowpass(data, f_upper, **kwargs),
            f_lower,
            **kwargs,
        )  # noqa
    else:
        raise ValueError(f"Invalid method '{method}'.")


def bessel_lowpass(data, fc, sample_rate, order=1, axis=-1):
    sos = sp.signal.bessel(
        2 * (order + 1),
        2 * fc / sample_rate,
        analog=False,
        btype="low",
        output="sos",
    )  # noqa
    return sp.signal.sosfilt(sos, data, axis=axis)


def bessel_highpass(data, fc, sample_rate, order=1, axis=-1):
    sos = sp.signal.bessel(
        2 * (order + 1),
        2 * fc / sample_rate,
        analog=False,
        btype="high",
        output="sos",
    )  # noqa
    return sp.signal.sosfilt(sos, data, axis=axis)



from __future__ import annotations

import glob
import logging
import os
from collections.abc import Mapping

import jax
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy as sp
from jax import scipy as jsp

from ..calibration import Calibration
from ..constants import MARIA_MAX_NU_HZ, MARIA_MIN_NU_HZ, c
from ..errors import FrequencyOutOfBoundsError
from ..spectrum import AtmosphericSpectrum
from ..units import Quantity
from ..utils import flatten_config, read_yaml

here, this_filename = os.path.split(__file__)

logger = logging.getLogger("maria")

BAND_FIELD_FORMATS = pd.read_csv(f"{here}/format.csv", index_col=0)

BAND_CONFIGS = {}
for path in glob.glob(f"{here}/configs/*.yml"):
    tag = os.path.split(path)[1].split(".")[0]
    BAND_CONFIGS[tag] = read_yaml(path)

BAND_CONFIGS = flatten_config(BAND_CONFIGS)

band_data = pd.DataFrame(BAND_CONFIGS).T.sort_index()
all_bands = list(band_data.index)


def parse_band(band):
    if isinstance(band, Band):
        return band
    if isinstance(band, Mapping):
        return Band(**band)
    if isinstance(band, str):
        return get_band(band)


def validate_band_config(band):
    if "passband" not in band:
        if any([key not in band for key in ["center", "width"]]):
            raise ValueError("The band's center and width must be specified")


def get_band(band_name):
    if band_name in BAND_CONFIGS:
        return Band(name=band_name, **BAND_CONFIGS[band_name])
    else:
        raise ValueError(f"'{band_name}' is not a valid pre-defined band name.")


def generate_passband(center, width, shape, samples=256):
    if shape == "flat":
        nu_min, nu_max = (center - 0.6 * width, center + 0.6 * width)
    elif shape == "top_hat":
        nu_min, nu_max = (center - width, center + width)
    else:
        nu_min, nu_max = (center - 1.5 * width, center + 1.5 * width)

    nu = np.linspace(nu_min, nu_max, samples)

    if shape == "flat":
        tau = np.where((nu > center - 0.5 * width) & (nu < center + 0.5 * width), 1, 0)
    elif shape == "gaussian":
        tau = np.exp(np.log(0.5) * (2 * (nu - center) / width) ** 2)
    elif shape == "top_hat":
        tau = np.exp(np.log(0.5) * (2 * (nu - center) / width) ** 8)
    else:
        raise ValueError(f"Invalid shape '{shape}'")

    if np.trapezoid(tau, x=nu) < 1e-2 * (nu_max - nu_min):
        raise ValueError("Error generating band")

    return nu, tau


class Band:
    def __init__(
        self,
        center: float = None,
        width: float = None,
        nu: float = None,
        tau: float = None,
        name: str = None,
        shape: str = "top_hat",
        efficiency: float = 0.5,
        sensitivity: float = None,
        NET_RJ: float = None,
        NET_CMB: float = None,
        NEP: float = None,
        NEP_per_loading: float = 0.0,
        gain_error: float = 0,
        knee: float = 1.0,
        time_constant: float = 0.0,
        spectrum_kwargs: dict = {},
    ):
        auto = center is not None and width is not None
        manual = nu is not None and tau is not None

        if not auto ^ manual:
            raise ValueError(
                "You must pass either both 'center' and 'width' or both 'nu' and 'tau'.",
            )

        if auto:
            nu, self.tau = generate_passband(center, width, shape, samples=1024)
            self.nu = Quantity(nu, "Hz")

        if manual:
            tau_max = np.max(tau)
            efficiency *= tau_max

            self.nu = Quantity(nu, "Hz")
            self.tau = np.array(tau) / tau_max

            if (self.nu.ndim != 1) or (self.tau.ndim != 1) or (self.nu.shape != self.tau.shape):
                raise ValueError(
                    f"'nu' and 'tau' have mismatched shapes ({self.nu.shape} and {self.tau.shape}).",
                )

        bad_freqs = self.nu[(self.nu.Hz < MARIA_MIN_NU_HZ) | (self.nu.Hz > MARIA_MAX_NU_HZ)]
        if bad_freqs:
            if nu is None:
                raise FrequencyOutOfBoundsError(center_and_width=(center, width))
            else:
                raise FrequencyOutOfBoundsError(nu=bad_freqs)

        # this turns e.g. 56MHz to "f056" and 150GHz to "f150"
        self.name = name or f"f{10 ** (np.log10(self.center.Hz) % 3):>03.0f}"
        self.shape = shape
        self.efficiency = efficiency
        self.NEP_per_loading = Quantity(NEP_per_loading, "W√s")
        self.knee = knee
        self.time_constant = time_constant
        self.gain_error = gain_error

        self.spectrum_kwargs = {}
        if spectrum_kwargs:
            self.spectrum = AtmosphericSpectrum(region=spectrum_kwargs["region"])
            self.spectrum_kwargs["zenith_pwv"] = spectrum_kwargs.get("pwv", 1e0)
            self.spectrum_kwargs["base_temperature"] = spectrum_kwargs.get(
                "temperature", self.spectrum.side_base_temperature.mean()
            )
            self.spectrum_kwargs["elevation"] = np.radians(spectrum_kwargs.get("elevation", 45))
        else:
            self.spectrum = None

        if sensitivity:
            logger.warning(
                "The 'sensitivity' keyword is deprecated and will be removed in future releases. "
                "To specify noise levels in terms of sky temperature, use the 'NET_RJ' or 'NET_CMB' keywords instead."
            )
            NET_RJ = sensitivity

        if NEP is not None:
            self.NEP = Quantity(NEP, "W√s")
        elif NET_RJ is not None:
            self.NET_RJ = NET_RJ
        elif NET_CMB is not None:
            self.NET_CMB = NET_CMB
        else:
            logger.warning(f"No noise level specified for band {self.name}, assuming a sensitivity of 50 uK_RJ√s.")
            self.NET_RJ = 50e-6

        self.transmission_integral_grids = {}

    @property
    def center(self):
        return Quantity(np.round(np.sum(self.nu.Hz * self.tau) / np.sum(self.tau), 2), "Hz")

    @property
    def width(self):
        crossovers = np.where((self.tau[1:] > 0.5) != (self.tau[:-1] > 0.5))[0]
        return Quantity(
            np.ptp(
                [sp.interpolate.interp1d(self.tau[[i, i + 1]], self.nu.Hz[[i, i + 1]])(0.5) for i in crossovers],
            ),
            "Hz",
        )

    def summary(self):
        filling = {
            "name": self.name,
            "center": self.center,
            "width": self.width,
            "η": self.efficiency,
            "NEP": self.NEP,
            "NET_RJ": self.NET_RJ,
            "NET_CMB": self.NET_CMB,
        }

        summary = pd.Series(filling)
        return summary

    def __repr__(self):
        return f"Band({', '.join([f'{index}={entry}' for index, entry in self.summary().items()])})"

    def plot(self):
        fig, ax = plt.subplots(1, 1)
        ax.plot(self.nu.Hz, self.tau, label=self.name)
        ax.set_xlim(self.nu.Hz.min(), self.nu.Hz.max())
        ax.set_xlabel(rf"$\nu$ [${self.nu.u['math_name']}$]")
        ax.set_ylabel(r"$\tau(\nu)$ [Rayleigh-Jeans]")
        ax.legend()

    @property
    def NET_RJ(self):
        return Quantity(self.cal("W -> K_RJ", spectrum=self.spectrum, **self.spectrum_kwargs)(self.NEP.to("W√s")), "K_RJ√s")

    @NET_RJ.setter
    def NET_RJ(self, value):
        self.NEP = Quantity(self.cal("K_RJ -> W", spectrum=self.spectrum, **self.spectrum_kwargs)(value), "W√s")

    @property
    def NET_CMB(self):
        return Quantity(
            self.cal("W -> K_CMB", spectrum=self.spectrum, **self.spectrum_kwargs)(self.NEP.to("W√s")), "K_CMB√s"
        )

    @NET_CMB.setter
    def NET_CMB(self, value):
        self.NEP = Quantity(self.cal("K_CMB -> W", spectrum=self.spectrum, **self.spectrum_kwargs)(value), "W√s")

    def compute_nu_integral(
        self,
        spectrum: AtmosphericSpectrum = None,
        nu_min_Hz: float = 0,
        nu_max_Hz: float = np.inf,
        **kwargs,
    ):
        """
        This is only useful for Rayleigh-Jeans sources (i.e. with linear emission).
        """

        if spectrum is None:
            nu = self.nu.Hz[(self.nu.Hz >= nu_min_Hz) & (self.nu.Hz < nu_max_Hz)]
            return np.trapezoid(y=self.passband(nu), x=nu, axis=-1)

        else:
            nu_mask = (spectrum.side_nu >= nu_min_Hz) & (spectrum.side_nu < nu_max_Hz)
            nu = spectrum.side_nu[nu_mask]
            integral_grid = np.trapezoid(y=self.passband(nu) * np.exp(-spectrum._opacity[..., nu_mask]), x=nu, axis=-1)
            xi = (kwargs["base_temperature"], kwargs["zenith_pwv"], kwargs["elevation"])
            return np.array(jsp.interpolate.RegularGridInterpolator(points=spectrum.points[:3], values=integral_grid)(xi))

    def transmission(self, region="chajnantor", pwv=1, elevation=np.radians(90)) -> float:
        if not hasattr(self, "spectrum"):
            self.spectrum = AtmosphericSpectrum(region=region)
        elif self.spectrum.region != region:
            self.spectrum = AtmosphericSpectrum(region=region)
        return self.spectrum.transmission(nu=self.center, pwv=pwv, elevation=elevation)

    # @classmethod
    # def from_file(cls, filename, **kwargs):
    #     path = (
    #         f"{here}/{filename}" if os.path.exists(f"{here}/{filename}") else filename
    #     )

    #     df = pd.read_csv(path, index_col=0)

    #     nu, tau = df.nu.values, df.passband.values

    #     # dummy values for center and width
    #     band = cls(nu=nu, tau=tau, shape="custom", **kwargs)
    #     band.nu = nu
    #     band.tau = tau

    def passband(self, nu):
        return self.efficiency * sp.interpolate.interp1d(
            self.nu.Hz,
            self.tau,
            bounds_error=False,
            fill_value=0,
        )(nu)

    @property
    def wavelength(self):
        """
        Return the wavelength of the center, in meters.
        """
        return c / self.center.Hz

    def cal(self, signature: str, **kwargs) -> float:
        return Calibration(
            signature,
            band=self,
            **kwargs,
        )



import logging
import os
import shutil
import time as ttime

import numpy as np
import requests
from requests import HTTPError
from tqdm import tqdm

from ..constants import DEFAULT_BAR_FORMAT
from ..utils.io import test_file

logger = logging.getLogger("maria")
here, this_filename = os.path.split(__file__)


def set_cache_dir(directory):
    os.environ["MARIA_CACHE_DIR"] = directory


def copy_file(source, destination):
    dest_dir, _ = os.path.split(destination)
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir, exist_ok=True)
    shutil.copy(source, destination)


def cache_status(path: str, max_age: float = 30 * 86400, refresh: bool = False):
    """
    Check if we need to reload the cache.
    """
    if refresh:
        logger.debug(f"Forcing refresh of {path}.")
        return "force_refresh"

    if not os.path.exists(path):
        logger.debug(f"Cached file at {path} does not exist.")
        return "missing"

    if not test_file(path):
        logger.debug(f"Could not open cached file at {path}.")
        return "corrupted"

    cache_age = ttime.time() - os.path.getmtime(path)

    if cache_age > max_age:
        logger.debug(f"Cached file at {path} is stale.")
        return "stale"

    return "ok"


def download_from_url(
    source_url: str,
    cache_path: str = None,
    chunk_size: int = 2**12,
    max_age: int = 30 * 86400,
):
    """
    Download the cache if needed.
    """
    cache_dir = os.path.dirname(cache_path)

    # make the cache directory if it doesn't exist
    if not os.path.exists(cache_dir):
        logger.debug(f"Creating {cache_dir}")
        os.makedirs(cache_dir, exist_ok=True)

    try:
        with requests.get(source_url, stream=True) as r:
            r.raise_for_status()
            total_size_bytes = int(r.headers.get("content-length", 0))
            logger.info(f"Fetching {source_url}")
            with tqdm(
                total=total_size_bytes,
                unit="B",
                unit_scale=True,
                desc=f"Downloading",
                bar_format=DEFAULT_BAR_FORMAT,
                ncols=250,
            ) as pbar:
                with open(cache_path, "wb") as f:
                    for chunk in r.iter_content(chunk_size):
                        f.write(chunk)
                        pbar.update(len(chunk))

    except HTTPError as error:
        if error.response.status_code == 404:
            raise error
        return f"Encountered error while downloading {source_url}: {repr(error)}"

    return cache_status(cache_path, max_age=max_age, refresh=False)


def fetch(
    source_path: str,
    cache_path: str = None,
    max_age: float = 7 * 86400,
    refresh: bool = False,
    url_base: str = "https://github.com/thomaswmorris/maria-data/raw/master",
    max_attempts: int = 7,
    **download_kwargs,
):
    """
    Fetch a file from the repo.
    """

    cache_dir = os.environ.get("MARIA_CACHE_DIR", f"/tmp/maria-data")
    cache_path = cache_path or f"{cache_dir}/{source_path}"
    source_url = f"{url_base}/{source_path}"

    # do we need to do anything?
    status = cache_status(cache_path, max_age=max_age, refresh=refresh)

    if status == "ok":
        return cache_path

    # do we have a potential backup?
    if status == "stale":
        stale_cache_path = f"{cache_dir}/stale/{source_path}"
        copy_file(cache_path, stale_cache_path)
    else:
        stale_cache_path = None

    attempt = 0
    while attempt < max_attempts:
        status = download_from_url(source_url, cache_path=cache_path, max_age=max_age, **download_kwargs)
        if status == "ok":
            return cache_path
        attempt += 1
        logger.warning(f"Could not download {source_url} on try {attempt} (status = {status})")
        ttime.sleep(2e0)

    if stale_cache_path:
        logger.warning(f"Could not download {source_url}, using stale cache at {stale_cache_path}")
        return stale_cache_path

    raise RuntimeError(f"Could not download {source_url} after {max_attempts} retries (status = {status})")



from __future__ import annotations

import numpy as np
import pytest

import maria
from maria.weather import Weather


@pytest.mark.parametrize("region_name", maria.all_regions)
def test_weather_from_cache(region_name):
    weather = Weather(region=region_name)

    print(f"{weather.pwv =}")


@pytest.mark.parametrize("region_name", ["chajnantor"])
def test_weather_from_cache_refresh(region_name):
    weather = Weather(region=region_name, refresh_cache=True)  # noqa

    h = weather.base_altitude + np.linspace(0, 1000, 10)

    for k, v in weather(h).items():
        assert not np.isnan(v).any()



# try:
#     import torch
# except ImportError:
#     raise ImportError("To do maximum likelihood mapping in maria, please install PyTorch (pip install torch)")

# import matplotlib.pyplot as plt
# import numpy as np
# import scipy as sp
# import torch
# from tqdm import tqdm, trange

# torch.set_num_threads(1)

# from __future__ import annotations

# import os
# from collections.abc import Sequence
# from time import sleep

# import dask.array as da
# import numpy as np
# import scipy as sp
# from torch import nn
# from tqdm import tqdm

# from maria.utils.signal import fit_bspline

# from ..band import Band, BandList
# from ..map import ProjectedMap
# from ..tod import TOD
# from .base import BaseProjectionMapper


# def remove_ends(data, ntip=16):
#     start_values = data[..., :ntip].mean(axis=-1)
#     end_values = data[..., -ntip:].mean(axis=-1)

#     return data - torch.linspace(start_values, end_values, data.shape[-1]).T


# class MaximumLikelihoodMapper(BaseProjectionMapper, nn.Module):
#     def __init__(
#         self,
#         center: tuple[float, float] = None,
#         stokes: str = "I",
#         width: float = 1,
#         height: float = None,
#         resolution: float = 0.01,
#         frame: str = "ra/dec",
#         units: str = "K_RJ",
#         degrees: bool = True,
#         calibrate: bool = False,
#         tod_preprocessing: dict = {},
#         map_postprocessing: dict = {},
#         tods: Sequence[TOD] = [],
#     ):
#         super(MaximumLikelihoodMapper, self).__init__()

#         height = height or width

#         super().__init__(
#             center=center,
#             stokes=stokes,
#             width=width,
#             height=height,
#             resolution=resolution,
#             frame=frame,
#             degrees=degrees,
#             calibrate=calibrate,
#             tods=tods,
#             units=units,
#             tod_preprocessing=tod_preprocessing,
#             map_postprocessing=map_postprocessing,
#         )

#         self.map = map
#         self.units = units

#         self.tods = [tod.to(self.units) for tod in tods]
#         self.tod_list = {}

#         for tod_index, tod in tqdm(enumerate(self.tods)):
#             t = {"tod": tod}
#             self.tod_list[tod_index] = t
#             ntip = int(0.5 * t["tod"].sample_rate)
#             d = tod.signal.compute()
#             d = d - fit_bspline(d.mean(axis=0), x=tod.time, spacing=10)
#             d = d - np.linspace(d[:, :ntip].mean(axis=-1), d[:, -ntip:].mean(axis=-1), d.shape[-1]).T
#             t["d"] = torch.tensor(d, dtype=torch.float)
#             t["w"] = torch.tensor(sp.signal.windows.hamming(M=t["d"].shape[-1]), dtype=torch.float)
#             t["f"] = np.fft.fftfreq(n=t["d"].shape[-1], d=1 / t["tod"].sample_rate)
#             t["abs_f"] = np.abs(t["f"])

#             idx, cum_npix = self.map._pointing_matrix_sparse_indices(t["tod"].coords)
#             indices = torch.stack([torch.arange(len(idx), dtype=torch.long), torch.tensor(idx, dtype=torch.long)], dim=0)
#             t["P"] = torch.sparse_coo_tensor(
#                 indices=indices, values=torch.ones(len(idx), dtype=torch.float), size=(len(idx), cum_npix)
#             )

#         self.sol_list = []
#         self.iteration = 0
#         self.iter_maps = []

#         self.norm = 1

#         self.initialize_mapper()

#         self.tods = []
#         self.add_tods(tods)

#     def initialize_mapper(self):
#         self.bands = BandList(bands=[])

#         for tod_index, tod in tqdm(enumerate(self.tods)):
#             t = {"tod": tod}
#             self.tod_list[tod_index] = t
#             ntip = int(0.5 * t["tod"].sample_rate)
#             d = d - np.linspace(d[:, :ntip].mean(axis=-1), d[:, -ntip:].mean(axis=-1), d.shape[-1]).T
#             t["d"] = torch.tensor(d, dtype=torch.float)
#             t["w"] = torch.tensor(sp.signal.windows.hamming(M=t["d"].shape[-1]), dtype=torch.float)
#             t["f"] = np.fft.fftfreq(n=t["d"].shape[-1], d=1 / t["tod"].sample_rate)
#             t["abs_f"] = np.abs(t["f"])

#             # compute the pointing tensor
#             idx, cum_npix = self.map._pointing_matrix_sparse_indices(t["tod"].coords)
#             indices = torch.stack([torch.arange(len(idx), dtype=torch.long), torch.tensor(idx, dtype=torch.long)], dim=0)
#             t["P"] = torch.sparse_coo_tensor(
#                 indices=indices, values=torch.ones(len(idx), dtype=torch.float), size=(len(idx), cum_npix)
#             )

#         self.iter_sol = nn.Parameter(data=self.naive_map().ravel().clone())

#         self.map = ProjectedMap(
#             data=np.zeros(self.map_shape),
#             weight=np.ones(self.map_shape),
#             stokes=self.stokes,
#             nu=self.bands.center,
#             resolution=self.resolution,
#             center=self.center,
#             degrees=False,
#             frame=self.frame.name,
#             units=self.units,
#         )

#         self.start_iteration()

#     def solution(self):
#         return sum(self.iter_maps) + self.m().detach()

#     def start_iteration(self):
#         self.iteration += 1
#         self.iter_maps.append(self.iter_sol.detach().reshape(self.map_shape()))
#         self.iter_sol = nn.Parameter(data=torch.zeros(len(self.iter_sol)))

#         for tod_index in self.tod_list:
#             t = self.tod_list[tod_index]

#             # project the fitted map out of the data
#             t["d_no_map"] = t["d"] - (t["P"] @ sum(self.iter_maps).ravel()).reshape(t["d"].shape)

#             t = self.tod_list[tod_index]

#             fd = torch.fft.fft(t["d_no_map"] * t["w"])
#             t["ps"] = fd.square().abs().float()

#             n_bin = 128
#             mid_f = np.geomspace(t["abs_f"][t["abs_f"] > 0].min(), t["abs_f"].max(), n_bin)
#             dlogf = np.exp(np.gradient(np.log(mid_f)).mean())
#             bin_f = np.geomspace(mid_f[0] / np.sqrt(dlogf), mid_f[-1] * np.sqrt(dlogf), n_bin + 1)
#             bin_y = sp.stats.binned_statistic(t["abs_f"], t["ps"].log(), bins=bin_f).statistic

#             use = ~np.isnan(bin_y).any(axis=0)
#             t["int_ps"] = torch.tensor(
#                 sp.interpolate.interp1d(mid_f[use], bin_y[:, use], axis=1, bounds_error=False, fill_value="extrapolate")(
#                     t["abs_f"]
#                 ),
#                 dtype=torch.float,
#             ).exp()

#             t["PNd"] = t["P"].T @ self.apply_inverse_noise_covariance(t["d_no_map"], tod_index=tod_index)

#         # self.sol = nn.Parameter(torch.zeros_like(self.naive_map().ravel()))

#         # self.loss_norm = 1
#         # self.loss_norm = self.loss().item()

#         fs = []
#         for tod_index, t in self.tod_list.items():
#             f = t["P"].T @ self.apply_inverse_noise_covariance(
#                 (t["P"] @ self.naive_map().ravel()).reshape(t["d"].shape), tod_index=tod_index
#             )
#             fs.append(f)

#         self.norm = 1 / torch.tensor([f.std() for f in fs]).mean()

#     def apply_inverse_noise_covariance(self, d, tod_index=0):
#         t = self.tod_list[tod_index]
#         return torch.fft.ifft(torch.fft.fft(t["w"] * d, axis=-1) / t["int_ps"]).real.ravel()

#     def map_weight_list(self, index=None):
#         return [t["P"].sum(axis=0).to_dense().reshape(self.map_shape()) for tod_index, t in self.tod_list.items()]

#     def map_weight(self, index=None):
#         return sum(self.map_weight_list())

#     def naive_map(self):
#         cum_wm = 0
#         cum_w = 0
#         map_weight_list = self.map_weight_list()
#         for tod_index, t in self.tod_list.items():
#             cum_wm += (t["P"].T @ t["d"].ravel()).reshape(self.map_shape())
#             cum_w += map_weight_list[tod_index]
#         return cum_wm / cum_w

#     def m(self):
#         return self.iter_sol.reshape(self.map.dims["y"] + 2, self.map.dims["x"] + 2)

#     def forward(self, tod_index: int = None):
#         t = self.tod_list[tod_index]
#         return t["P"].T @ self.apply_inverse_noise_covariance(
#             (t["P"] @ self.iter_sol).reshape(t["d"].shape), tod_index=tod_index
#         )

#     def forward_list(self):
#         return [self.forward(tod_index) for tod_index in self.tod_list]
#         # if index is None:
#         #     return [t["P"].T @ self.apply_inverse_noise_covariance((t["P"]
#  @ self.sol).reshape(t["d"].shape), tod_index=tod_index)
#         #

#     def loss(self):
#         return sum(
#             [
#                 self.norm**2 * (PNPm - t["PNd"]).square().mean()
#                 for PNPm, t in zip(self.forward_list(), self.tod_list.values())
#             ]
#         )


# output_map = input_map[..., :, :]

# ml_mapper = MLMapper(tods=tods, map=output_map, units="mK_RJ")
# # bpm#.loss()
# self = ml_mapper.to("cpu")
# spectra = {}
# self.loss()



from __future__ import annotations

import numpy as np
import scipy as sp  # noqa F401

from ..constants import c


def compute_angular_fwhm(fwhm_0, z=np.inf, n=1.0, nu=None, l=None):  # noqa F401
    """
    Returns the angular full width at half maximum of a Gaussian beam at distance `z` in
    refractive index `n`. Supply either the wavelength `l` in meters or the frequency `nu` in GHz.

    NOTE: For telescope purposes, `fwhm_0` is the width of the objective/primary.
    """

    if (nu is None) and (l is None):
        raise ValueError("You must supply either a frequency 'f' or wavelength 'l'.")

    w_0 = fwhm_0 / 2

    # Rayleigh range
    z_r = np.pi * w_0**2 * n / (l or c / nu)

    return 2 * w_0 * np.sqrt(1 / z**2 + 1 / z_r**2)


def compute_physical_fwhm(fwhm_0, z=np.inf, n=1, nu=None, l=None):  # noqa F401
    return z * compute_angular_fwhm(fwhm_0=fwhm_0, z=z, n=n, nu=nu, l=l)


def construct_beam_filter(fwhm, res, beam_profile=None, buffer=1):
    """
    Make a beam filter for an image.
    """

    if beam_profile is None:
        # beam_profile = lambda r, r0: np.where(r <= r0, 1., 0.)

        # a top hat
        def beam_profile(r, r0):
            return np.exp(-((r / r0) ** 16))

    filter_width = buffer * fwhm

    n_side = int(np.maximum(filter_width / res, 3))

    filter_side = np.linspace(-filter_width / 2, filter_width / 2, n_side)
    X, Y = np.meshgrid(filter_side, filter_side, indexing="ij")
    R = np.sqrt(np.square(X) + np.square(Y))
    F = beam_profile(R, fwhm / 2)

    return F / F.sum()


def separably_filter_2d(data, F, tol=1e-2, return_filter=False):
    """
    This is more efficient than 2d convolution.
    """

    if F.ndim != 2:
        raise ValueError("'F' must be two-dimensional.")

    u, s, v = np.linalg.svd(F)
    effective_filter = 0
    filtered_image = 0

    for m in range(len(F)):
        effective_filter += s[m] * u[:, m : m + 1] @ v[m : m + 1]
        filtered_image += s[m] * sp.ndimage.convolve1d(
            sp.ndimage.convolve1d(data, u[:, m], axis=-2),
            v[m],
            axis=-1,
        )

        if np.abs(F - effective_filter).mean() < tol:
            break

    return (filtered_image, effective_filter) if return_filter else filtered_image



from __future__ import annotations

import gc
import logging
import os
import time as ttime

import arrow
import numpy as np

from ..coords import Coordinates
from ..instrument import Instrument, get_instrument
from ..io import humanize_time
from ..plan import Plan, PlanList, get_plan
from ..site import Site, get_site
from ..tod import TOD
from ..utils import read_yaml

here, this_filename = os.path.split(__file__)

logger = logging.getLogger("maria")


class InvalidSimulationParameterError(Exception):
    def __init__(self, invalid_keys):
        super().__init__(
            f"The parameters {invalid_keys} are not valid simulation parameters!",
        )


master_params = read_yaml(f"{here}/params.yml")


def parse_sim_kwargs(kwargs, master_kwargs, strict=False):
    parsed_kwargs = {k: {} for k in master_kwargs.keys()}
    invalid_kwargs = {}

    for k, v in kwargs.items():
        parsed = False
        for sub_type, sub_kwargs in master_kwargs.items():
            if k in sub_kwargs.keys():
                parsed_kwargs[sub_type][k] = v
                parsed = True
        if not parsed:
            invalid_kwargs[k] = v

    if len(invalid_kwargs) > 0:
        if strict:
            raise InvalidSimulationParameterError(
                invalid_keys=list(invalid_kwargs.keys()),
            )

    return parsed_kwargs


class BaseSimulation:
    """
    The base class for a simulation. This is an ingredient in every simulation.
    """

    def __init__(
        self,
        instrument: Instrument | str,
        plan: Plan | str,
        site: Site | str,
        progress_bars: bool = True,
        dtype=np.float32,
        **kwargs,
    ):
        start_init_s = ttime.monotonic()

        if hasattr(self, "boresight"):
            return

        self.dtype = dtype
        self.disable_progress_bars = (not progress_bars) or (logging.getLevelName(logger.level) == "DEBUG")
        parsed_sim_kwargs = parse_sim_kwargs(kwargs, master_params)

        if isinstance(instrument, Instrument):
            self.instrument = instrument
        else:
            self.instrument = get_instrument(
                name=instrument,
                **parsed_sim_kwargs["instrument"],
            )

        logger.debug(f"Initialized instrument in {humanize_time(ttime.monotonic() - start_init_s)}.")
        instrument_init_s = ttime.monotonic()

        # if not isinstance(plan, list):
        #     plan = [plan]

        if isinstance(plan, Plan):
            self.plan = plan
        else:
            self.plan = get_plan(plan_name=plan, **parsed_sim_kwargs["plan"])

        self.plans = PlanList([plan])

        logger.debug(f"Initialized plan in {humanize_time(ttime.monotonic() - instrument_init_s)}.")
        plan_init_s = ttime.monotonic()

        if isinstance(site, Site):
            self.site = site
        elif isinstance(site, str):
            self.site = get_site(site_name=site, **parsed_sim_kwargs["site"])
        else:
            raise ValueError(
                "'site' must be either a Site object or a string.",
            )

        logger.debug(f"Initialized site in {humanize_time(ttime.monotonic() - plan_init_s)}.")
        site_init_s = ttime.monotonic()

        self.boresight = Coordinates(
            t=self.plan.time,
            phi=self.plan.phi,
            theta=self.plan.theta,
            earth_location=self.site.earth_location,
            frame=self.plan.frame,
        )

        logger.debug(f"Initialized boresight in {humanize_time(ttime.monotonic() - site_init_s)}.")
        boresight_init_s = ttime.monotonic()

        if self.plan.max_vel_deg > self.instrument.vel_limit:
            raise ValueError(
                (
                    f"The maximum velocity of the boresight ({self.plan.max_vel_deg:.01f} deg/s) exceeds "
                    f"the maximum velocity of the instrument ({self.instrument.vel_limit:.01f} deg/s)."
                ),
            )

        if self.plan.max_acc_deg > self.instrument.acc_limit:
            raise ValueError(
                (
                    f"The maximum acceleration of the boresight ({self.plan.max_acc_deg:.01f} deg/s^2) exceeds "
                    f"the maximum acceleration of the instrument ({self.instrument.acc_limit:.01f} deg/s^2)."
                ),
            )

        # this can be expensive sometimes
        self.coords = self.boresight.broadcast(
            self.instrument.dets.offsets,
            frame="az/el",
        )

        logger.debug(f"Initialized coordinates in {humanize_time(ttime.monotonic() - boresight_init_s)}.")

        logger.debug(f"Initialized generic simulation in {humanize_time(ttime.monotonic() - start_init_s)}.")

    @property
    def shape(self):
        return (self.instrument.n, self.plan.n)

    def _run(self):
        raise NotImplementedError()

    def run(self, units: str = "K_RJ"):
        group_index_list = self.plans.plan_groups()

        for plan, plan_index in zip(self.plans.group_plans(), group_index_list):
            logger.info(f"Simulating TODs {plan_index}")

            self.loading = {}
            self._run()

            metadata = {
                "atmosphere": False,
                "sim_time": arrow.now(),
                "altitude": float(self.site.altitude.m),
                "region": self.site.region,
            }

        if hasattr(self, "atmosphere"):
            metadata["atmosphere"] = True
            metadata["pwv"] = float(np.round(self.atmosphere.weather.pwv, 3))
            metadata["base_temperature"] = float(np.round(self.atmosphere.weather.temperature[0], 3))

        tod = TOD(
            data=self.loading,
            dets=self.instrument.dets,
            coords=self.coords,
            units="pW",
            metadata=metadata,
        )

        gc.collect()

        return tod.to(units)

    def plot_counts(self, x_bins=100, y_bins=100):
        self.plan.plot_counts(instrument=self.instrument, x_bins=x_bins, y_bins=y_bins)



import numpy as np
import scipy as sp

from ..constants import T_CMB, k_B
from ..errors import ShapeError
from ..functions.radiometry import (
    inverse_planck_spectrum,
    inverse_rayleigh_jeans_spectrum,
    planck_spectrum,
    rayleigh_jeans_spectrum,
)  # noqa


def identity(x: float, **kwargs):
    return x


def cmb_temperature_anisotropy_to_brightness_temperature(dT_CMB, **kwargs):
    return dT_CMB + T_CMB


def brightness_temperature_to_cmb_temperature_anisotropy(T_b, **kwargs):
    return T_b - T_CMB


def rayleigh_jeans_temperature_to_brightness_temperature(T_RJ, nu, **kwargs):
    I_nu = rayleigh_jeans_spectrum(T_RJ=T_RJ, nu=nu)
    return inverse_planck_spectrum(I_nu=I_nu, nu=nu)


def brightness_temperature_to_rayleigh_jeans_temperature(T_b, nu, **kwargs):
    I_nu = planck_spectrum(T_b=T_b, nu=nu)
    return inverse_rayleigh_jeans_spectrum(I_nu=I_nu, nu=nu)


def rayleigh_jeans_temperature_to_radiant_flux(
    T_RJ,
    band,
    polarized: bool = False,
    spectrum=None,
    **kwargs,
):
    """
    nu: frequency, in Hz
    band: response to a Rayleigh-Jeans source
    """
    if spectrum:
        integral = band.compute_nu_integral(
            spectrum=spectrum,
            zenith_pwv=kwargs["zenith_pwv"],
            base_temperature=kwargs["base_temperature"],
            elevation=kwargs["elevation"],
        )

    else:
        integral = band.compute_nu_integral(spectrum=None)

    return (0.5 if polarized else 1.0) * k_B * integral * T_RJ


def radiant_flux_to_rayleigh_jeans_temperature(P, band, polarized: bool = False, spectrum=None, **kwargs):
    """
    nu: frequency, in Hz
    passband: response to a Rayleigh-Jeans source
    """

    if spectrum:
        integral = band.compute_nu_integral(
            spectrum=spectrum,
            zenith_pwv=kwargs["zenith_pwv"],
            base_temperature=kwargs["base_temperature"],
            elevation=kwargs["elevation"],
        )

    else:
        integral = band.compute_nu_integral(spectrum=None)

    return P / ((0.5 if polarized else 1.0) * k_B * integral)


def brightness_temperature_to_radiant_flux_explicit(
    T_b, band, polarized: bool = False, spectrum=None, eps: float = 1e-4, **kwargs
):
    if T_b.ndim > 1:
        raise ShapeError("'T_b' must be one-dimensional")

    # stokes_factor = 0.5 if polarized else 1.0
    shaped_T_b = np.expand_dims(T_b, axis=0)

    if spectrum:
        # add dimensions to broadcast with the atmospheric spectrum
        shaped_T_b = np.expand_dims(shaped_T_b, axis=(0, 1, 2))
        shaped_nu = np.expand_dims(spectrum.side_nu, -1)
        sample_T_RJ = inverse_rayleigh_jeans_spectrum(
            planck_spectrum(T_b=shaped_T_b, nu=shaped_nu),
            nu=shaped_nu,
        )

        # this has shape (spectrum.shape, T_b.size)
        integral_samples = np.trapezoid(
            y=sample_T_RJ * (np.exp(-np.expand_dims(spectrum._opacity, -1)) * band.passband(shaped_nu)),
            x=spectrum.side_nu,
            axis=-2,
        )

        points = spectrum.points[:3]
        xi = (
            kwargs["base_temperature"],
            kwargs["zenith_pwv"],
            kwargs["elevation"],
        )
        integral = sp.interpolate.RegularGridInterpolator(points, integral_samples)(xi)

    else:
        shaped_nu = np.expand_dims(band.nu.Hz, -1)
        sample_T_RJ = inverse_rayleigh_jeans_spectrum(
            planck_spectrum(T_b=shaped_T_b, nu=shaped_nu),
            nu=shaped_nu,
        )
        integral = np.trapezoid(y=sample_T_RJ * band.passband(shaped_nu), x=band.nu.Hz, axis=-2)

    return (0.5 if polarized else 1.0) * k_B * integral


def brightness_temperature_to_radiant_flux(T_b, band, polarized: bool = False, spectrum=None, eps: float = 1e-4, **kwargs):
    T_b_lo = np.min(T_b) - eps / 2
    T_b_hi = np.min(T_b) + eps / 2

    # this has shape (*kwargs.shape, 2)
    P = brightness_temperature_to_radiant_flux_explicit(
        T_b=np.array([T_b_lo, T_b_hi]), band=band, polarized=polarized, spectrum=spectrum, **kwargs
    )

    t = (T_b - T_b_lo) / eps
    return t * P[..., 1] + (1 - t) * P[..., 0]


def radiant_flux_to_brightness_temperature(P, **kwargs):
    raise NotImplementedError()


def rayleigh_jeans_temperature_to_spectral_flux_density_per_pixel(T_RJ: float, nu: float, pixel_area: float, **kwargs):
    """
    T_RJ: Rayleigh-Jeans temperature, in Kelvin
    nu: frequency, in Hz
    res: resolution, in radians
    """
    return rayleigh_jeans_spectrum(T_RJ=T_RJ, nu=nu) * pixel_area


def spectral_flux_density_per_pixel_to_rayleigh_jeans_temperature(E: float, nu: float, pixel_area: float, **kwargs):
    """
    T_RJ: Rayleigh-Jeans temperature, in Jy/pixel
    nu: frequency, in Hz
    res: resolution, in radians
    """
    I_nu = E / pixel_area
    return inverse_rayleigh_jeans_spectrum(I_nu=I_nu, nu=nu)


def dP_dT_CMB(band, polarized=False, spectrum=None, eps=1e-4, **kwargs):
    sample_T_b = np.array([T_CMB - eps / 2, T_CMB + eps / 2])
    P = brightness_temperature_to_radiant_flux_explicit(
        T_b=sample_T_b, band=band, polarized=polarized, spectrum=spectrum, **kwargs
    )
    return (P[..., 1] - P[..., 0]) / eps


def cmb_temperature_anisotropy_to_radiant_flux(T_b, band, polarized=False, spectrum=None, **kwargs):
    return T_b * dP_dT_CMB(band=band, polarized=polarized, spectrum=spectrum, **kwargs)


def radiant_flux_to_cmb_temperature_anisotropy(P, band, polarized=False, spectrum=None, **kwargs):
    return P / dP_dT_CMB(band=band, polarized=polarized, spectrum=spectrum, **kwargs)


def T_RJ_per_T_CMB(
    band,
    eps: float = 1e-3,
    **kwargs,
):
    """
    Color correction for NO ATMOSPHERE
    """
    test_T_b = T_CMB + np.array([[-eps / 2], [+eps / 2]])
    T_RJ = inverse_rayleigh_jeans_spectrum(
        planck_spectrum(T_b=test_T_b, nu=band.nu.Hz),
        nu=band.nu.Hz,
    )
    P = k_B * np.trapezoid(T_RJ * band.passband(band.nu.Hz), x=band.nu.Hz, axis=-1)
    return radiant_flux_to_rayleigh_jeans_temperature((P[1] - P[0]) / eps, spectrum=None, band=band)


def cmb_temperature_anisotropy_to_rayleigh_jeans_temperature(
    delta_T_CMB: float,
    band,
    **kwargs,
):
    return delta_T_CMB * T_RJ_per_T_CMB(band=band)


def rayleigh_jeans_temperature_to_cmb_temperature_anisotropy(
    T_RJ: float,
    band,
    **kwargs,
):
    return T_RJ / T_RJ_per_T_CMB(band=band)


def spectral_radiance_to_spectral_flux_density_per_pixel(E: float, pixel_area: float, **kwargs):
    """
    E: Spectral flux density per area, in Jy/sr
    pixel_area: pixel area, in steradians
    """
    return E * pixel_area


def spectral_flux_density_per_pixel_to_spectral_radiance(E: float, pixel_area: float, **kwargs):
    """
    E: Spectral flux density per pixel, in Jy/pixel
    pixel_area: pixel area, in steradians
    """
    return E / pixel_area


def spectral_flux_density_per_beam_to_spectral_flux_density_per_pixel(
    E: float, pixel_area: float, beam_area: float, **kwargs
):
    """
    E: Spectral flux density per area, in Jy/sr
    pixel_area: pixel area, in steradians
    """
    return E * pixel_area / beam_area


def spectral_flux_density_per_pixel_to_spectral_flux_density_per_beam(
    E: float, pixel_area: float, beam_area: float, **kwargs
):
    """
    E: Spectral flux density per pixel, in Jy/pixel
    pixel_area: pixel area, in steradians
    """
    return E * beam_area / pixel_area



from __future__ import annotations

import matplotlib.pyplot as plt
import pytest

import maria

plt.close("all")


@pytest.mark.parametrize("site_name", maria.all_sites)
def test_get_site(site_name):
    site = maria.get_site(site_name)
    print(site)


def test_site_plot():
    site = maria.get_site("cerro_toco")
    site.plot()



from __future__ import annotations

import os

import dask.array as da
import numpy as np
from tqdm import tqdm

from ..io import DEFAULT_BAR_FORMAT
from ..noise import generate_noise_with_knee


class NoiseMixin:
    def _run(self):
        self._simulate_noise()

    def _simulate_noise(self, obs):
        noise_loading = np.zeros(shape=obs.shape, dtype=self.dtype)

        bands_pbar = tqdm(
            obs.instrument.dets.bands,
            desc="Generating noise",
            disable=self.disable_progress_bars,
            bar_format=DEFAULT_BAR_FORMAT,
            ncols=250,
        )
        for band in bands_pbar:
            bands_pbar.set_postfix({"band": band.name})

            band_mask = obs.instrument.dets.band_name == band.name

            total_NEP = band.NEP.to("W√s") + band.NEP_per_loading.to("W√s") * sum(
                [d[band_mask].compute() for d in obs.loading.values()]
            )

            noise_loading[band_mask] = (
                1e12
                * total_NEP
                * generate_noise_with_knee(
                    shape=(band_mask.sum(), obs.plan.n),
                    sample_rate=obs.plan.sample_rate.Hz,
                    knee=band.knee,
                )
            )

        obs.loading["noise"] = da.asarray(noise_loading, dtype=self.dtype)



from __future__ import annotations

import logging
import os
from pathlib import Path

import arrow
import numpy as np
import pandas as pd
import scipy as sp
from astropy.coordinates import EarthLocation
from astropy.io import fits
from astropy.wcs import WCS
from matplotlib import pyplot as plt

from .. import coords
from ..coords import Coordinates, Frame, frames, get_center_phi_theta, phi_theta_to_offsets
from ..io import DEFAULT_TIME_FORMAT, repr_lat_lon, repr_phi_theta
from ..site import Site, get_site
from ..units import Quantity
from ..utils import compute_diameter, read_yaml
from .patterns import get_scan_pattern_generator, scan_patterns

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")

all_patterns = list(scan_patterns.index.values)

MAX_VELOCITY_WARN = 10  # in deg/s
MAX_ACCELERATION_WARN = 10  # in deg/s

MIN_ELEVATION_WARN = 20  # in deg
MIN_ELEVATION_ERROR = 10  # in deg

here, this_filename = os.path.split(__file__)

PLAN_CONFIGS = {}
for plans_path in Path(f"{here}/plans").glob("*.yml"):
    PLAN_CONFIGS.update(read_yaml(plans_path))

plan_params = set()
for key, config in PLAN_CONFIGS.items():
    plan_params |= set(config.keys())
plan_data = pd.DataFrame(PLAN_CONFIGS).T
all_plans = list(plan_data.index.values)


class Plan:
    """
    A dataclass containing time-ordered plan data.
    """

    @classmethod
    def generate(
        cls,
        site: Site | str = None,
        description: str = "",
        start_time: str | int = None,
        duration: float = 60.0,
        sample_rate: float = 50.0,
        frame: str = "ra/dec",
        degrees: bool = True,
        jitter: float = 0.0,
        roll: float = 0.0,
        scan_center: tuple[float, float] = (0.0, 0.0),
        scan_pattern: str = "daisy",
        scan_options: dict = {},
    ):
        duration = Quantity(duration, "s")
        sample_rate = Quantity(sample_rate, "Hz")

        # for k, v in PLAN_CONFIGS[self.scan_pattern]["scan_options"].items():
        #     if k not in self.scan_options.keys():
        #         self.scan_options[k] = v

        if start_time is None:
            start_time = arrow.now().timestamp()
        start_time = arrow.get(start_time)

        time_min = start_time.timestamp()
        time_max = time_min + duration.seconds

        time = np.arange(time_min, time_max, 1 / sample_rate.Hz)

        # # convert radius to width / height
        # if "width" in self.scan_options:
        #     self.scan_options["radius"] = 0.5 * self.scan_options.pop("width")

        # this is in pointing_units
        scan_offsets = get_scan_pattern_generator(scan_pattern)(
            time,
            **scan_options,
        )

        assert not np.isnan(scan_offsets).any()

        scan_offsets = Quantity(scan_offsets, units=("deg" if degrees else "rad")).rad

        # TODO: scan speed checks in az/el frame

        # scan_velocity = np.gradient(
        #     self.scan_offsets,
        #     axis=1,
        #     edge_order=0,
        # ) / np.gradient(self.time)

        # scan_acceleration = np.gradient(
        #     scan_velocity,
        #     axis=1,
        #     edge_order=0,
        # ) / np.gradient(self.time)

        # self.max_vel_deg = np.degrees(np.sqrt(np.sum(scan_velocity**2, axis=0)).max())
        # self.max_acc_deg = np.degrees(np.sqrt(np.sum(scan_acceleration**2, axis=0)).max())

        # if self.max_vel_deg > MAX_VELOCITY_WARN:
        #     logger.warning(
        #         (
        #             f"The maximum velocity of the boresight ({self.max_vel_deg:.01f} deg/s) is "
        #             "physically unrealistic. If this is undesired, double-check the parameters for your scan strategy."
        #         ),
        #         stacklevel=2,
        #     )

        # if self.max_acc_deg > MAX_ACCELERATION_WARN:
        #     logger.warning(
        #         (
        #             f"The maximum acceleration of the boresight ({self.max_acc_deg:.01f} deg/s^2) is "
        #             "physically unrealistic. If this is undesired, double-check the parameters for your scan strategy."
        #         ),
        #         stacklevel=2,
        #     )

        if len(scan_center) == 2:
            units = "deg" if degrees else "rad"
            scan_center = (Quantity(scan_center[0], units=units), Quantity(scan_center[1], units=units))
        else:
            raise ValueError("'scan_center' must be a 2-tuple of numbers")

        scan_offsets += np.radians(jitter) * np.random.standard_normal(size=scan_offsets.shape)  # noqa

        pt = coords.offsets_to_phi_theta(
            scan_offsets.T,
            scan_center[0].rad,
            scan_center[1].rad,
        )

        self = cls(time, phi=pt[..., 0], theta=pt[..., 1], roll=roll, frame=frame, site=site)

        self.generation_kwargs = {"scan_pattern": scan_pattern, "scan_options": scan_options}

        return self

    def __init__(
        self,
        time: float,
        phi: float,
        theta: float,
        roll: float = 0.0,
        frame: str = "ra/dec",
        site: Site | str = None,
        latitude: float = None,  # in degrees
        longitude: float = None,  # in degrees
        altitude: float = 0,  # in meters
    ):
        if site is not None:
            if isinstance(site, str):
                site = get_site(site)
            elif not isinstance(site, Site):
                raise TypeError()

            self.site = site
            earth_location = site.earth_location

        elif latitude is not None and longitude is not None:
            # self.latitude = Quantity(latitude, "deg")
            # self.longitude = Quantity(longitude, "deg")
            # self.altitude = Quantity(altitude, "m")
            earth_location = EarthLocation.from_geodetic(
                lon=longitude,
                lat=latitude,
                height=altitude,
            )

        else:
            earth_location = None

        self.coords = Coordinates(
            phi=phi,
            theta=theta,
            t=time,
            frame=frame,
            earth_location=earth_location,
        )

        if self.frame.name == "ra/dec":
            self.ra = self.phi = phi
            self.dec = self.theta = theta
        elif self.frame.name == "az/el":
            self.az = self.phi = phi
            self.el = self.theta = theta
        else:
            raise ValueError("Not a valid pointing frame!")

        offsets = self.offsets()
        self.scan_speed = Quantity(
            np.sqrt(np.square(np.gradient(offsets, axis=0)).sum(axis=1)) / np.gradient(self.time), "rad/s"
        )

        self.roll = roll

    @property
    def n(self):
        return len(self.coords.t)

    @property
    def time(self):
        return self.coords.t

    @property
    def frame(self):
        return self.coords.frame

    @property
    def earth_location(self):
        return self.coords.earth_location

    @property
    def dt(self):
        return np.median(np.diff(self.coords.t))

    @property
    def sample_rate(self):
        return Quantity(1 / self.dt, "Hz")

    @property
    def duration(self):
        return Quantity(np.ptp(self.coords.t) + self.dt, "s")

    @property
    def start_time(self):
        return arrow.get(self.time[0])

    @property
    def end_time(self):
        return self.start_time.shift(seconds=self.duration.s)

    @property
    def naive(self):
        return self.earth_location is None

    def center(self, frame: str = None, center: tuple[float, float] = None):
        frame = Frame(frame or self.frame.name)
        if center is not None:
            cphi, ctheta = center
        else:
            cphi, ctheta = get_center_phi_theta(
                phi=getattr(self.coords, frame.phi_name), theta=getattr(self.coords, frame.theta_name)
            )
        return (Quantity(cphi, "rad"), Quantity(ctheta, "rad"))

    def offsets(self, frame: str = None):
        center = self.center(frame=frame or self.frame.name)
        phi_theta = np.stack([self.phi, self.theta], axis=-1)
        return phi_theta_to_offsets(phi_theta, float(center[0].rad), float(center[1].rad))

    def plot(self, plot_az_el: bool = None):
        two_panel = plot_az_el if plot_az_el is not None else (self.frame.name != "az/el" and not self.naive)

        q_offsets = Quantity(self.offsets(), "rad")
        center = self.center()

        header = fits.header.Header()
        header["CDELT1"] = -np.degrees(q_offsets.u["factor"])
        header["CDELT2"] = np.degrees(q_offsets.u["factor"])
        header["CRPIX1"] = 1
        header["CRPIX2"] = 1
        header["CTYPE1"] = "RA---SIN"
        header["CUNIT1"] = "deg     "
        header["CTYPE2"] = "DEC--SIN"
        header["CUNIT2"] = "deg     "
        header["RADESYS"] = "FK5     "
        header["CRVAL1"] = center[0].deg
        header["CRVAL2"] = center[1].deg

        wcs = WCS(header)

        fig = plt.figure(figsize=(10, 5) if two_panel else (5, 5), dpi=256, constrained_layout=True)
        ax = fig.add_subplot(111 if two_panel is None else 121, projection=wcs)

        cphi_repr, ctheta_repr = repr_phi_theta(center[0].rad, center[1].rad, frame=self.frame.name)

        ax.plot(q_offsets.value[:, 0], q_offsets.value[:, 1], lw=5e-1)
        ax.scatter(0, 0, c="r", marker="x", label=f"{cphi_repr}\n{ctheta_repr}")

        ax.legend(loc="upper right")

        ax.set_aspect("equal")

        ax.tick_params(axis="x", bottom=True, top=False)
        ax.tick_params(axis="y", left=True, right=False, rotation=90)

        ax2 = ax.secondary_xaxis("top")
        ax2.set_xlabel(rf"$\Delta \, \theta_x$ [${q_offsets.u['math_name']}$]")
        ax.set_xlabel(rf"{self.frame.phi_long_name}")
        ax.set_ylabel(rf"{self.frame.theta_long_name}")

        if not two_panel:
            ay2 = ax.secondary_yaxis("right")
            ay2.set_ylabel(rf"$\Delta \, \theta_y$ [${q_offsets.u['math_name']}$]")

        if two_panel:
            az = Quantity(np.unwrap(self.coords.az), "rad")
            el = Quantity(self.coords.el, "rad")

            ax = fig.add_subplot(122)

            start_ha = "right" if az[0] > az.mean() else "left"
            start_xytext = (10 if start_ha == "left" else -10, 0)

            end_ha = "right" if az[-1] > az.mean() else "left"
            end_xytext = (10 if end_ha == "left" else -10, 0)

            annotate_kwargs = dict(
                va="center",
                xycoords="data",
                textcoords="offset points",
                bbox={"boxstyle": "round,pad=0.3", "fc": "w", "ec": "k", "alpha": 0.5, "lw": 1},
            )

            ax.scatter(az[0].deg, el[0].deg, color="g", marker="+")
            ax.scatter(az[-1].deg, el[-1].deg, color="r", marker="+")
            ax.annotate(
                xy=(az[0].deg, el[0].deg),
                xytext=start_xytext,
                text=f"START ({self.repr_start_time})",
                c="g",
                ha=start_ha,
                **annotate_kwargs,
            )
            ax.annotate(
                xy=(az[-1].deg, el[-1].deg),
                xytext=end_xytext,
                text=f"END ({self.repr_end_time})",
                c="r",
                ha=end_ha,
                **annotate_kwargs,
            )

            ax.plot(az.deg, el.deg, lw=5e-1)

            ax.set_xlabel("Azimuth [degrees]")
            ax.set_ylabel("Elevation [degrees]")

            ax.yaxis.tick_right()
            ax.yaxis.set_label_position("right")

    def map_counts(self, instrument=None, x_bins=64, y_bins=64):
        array_offsets = np.zeros((1, 1, 2)) if instrument is None else instrument.offsets[:, None]

        OFFSETS = self.offsets()[None] + array_offsets

        xmin, ymin = OFFSETS.min(axis=(0, 1))
        xmax, ymax = OFFSETS.max(axis=(0, 1))

        if isinstance(x_bins, int):
            x_bins = np.linspace(xmin, max(xmax, xmin + 1e-6), x_bins + 1)
        if isinstance(y_bins, int):
            y_bins = np.linspace(ymin, max(ymax, ymin + 1e-6), y_bins + 1)

        bs = sp.stats.binned_statistic_2d(
            OFFSETS[..., 1].ravel(),
            OFFSETS[..., 0].ravel(),
            0,
            statistic="count",
            bins=(y_bins, x_bins),
        )

        return x_bins, y_bins, bs[0]

    def plot_counts(self, instrument=None, x_bins=64, y_bins=64):
        fig, ax = plt.subplots(1, 1, figsize=(5, 4))

        x, y, counts = self.map_counts(instrument=instrument, x_bins=x_bins, y_bins=y_bins)
        q_x = Quantity(x, "rad")
        q_y = Quantity(y, "rad")

        heatmap = ax.pcolormesh(q_x.value, q_y.value, counts, cmap="turbo", vmin=0)
        ax.set_xlabel(rf"$\Delta \theta_x$ [{q_x.u['math_name']}]")
        ax.set_ylabel(rf"$\Delta \theta_y$ [{q_y.u['math_name']}]")

        cbar = fig.colorbar(heatmap, location="right")
        cbar.set_label("counts")

    @property
    def repr_start_time(self):
        return self.start_time.format(DEFAULT_TIME_FORMAT)

    @property
    def repr_end_time(self):
        return self.end_time.format(DEFAULT_TIME_FORMAT)

    def __repr__(self):
        c = self.center(frame=self.frame.name)
        q_offsets = Quantity(self.offsets(), "rad")

        cphi_repr, ctheta_repr = repr_phi_theta(c[0].rad, c[1].rad, frame=self.frame.name)
        center_string = f"""center:
    {cphi_repr}
    {ctheta_repr}"""

        if not self.naive:
            repr_lat, repr_lon = repr_lat_lon(self.site.latitude.degrees, self.site.longitude.degrees)
            location_string = f"""
    lat: {repr_lat}
    lon: {repr_lon}
    alt: {self.site.altitude}"""
            if self.frame.name != "az/el":
                caz, cel = self.center(frame="az/el")
                center_string = f"""center:
    {cphi_repr}
    {ctheta_repr}
    az(mean): {caz}
    el(mean): {cel}"""
        else:
            location_string = f"naive"

        return f"""Plan:
  duration: {self.duration}
    start: {self.repr_start_time}
    end:   {self.repr_end_time}
  location: {location_string}
  sample_rate: {self.sample_rate}
  {center_string}
  scan_radius: {Quantity(compute_diameter(q_offsets.rad), "rad")}
  scan_speed(mean): {self.scan_speed.mean()}"""

    def __add__(self, other):
        if not isinstance(other, type(self)):
            raise TypeError("Plans can only be added to other Plans")

        assert self.end_time < other.start_time

        time = np.r_[self.time, other.time]
        phi = np.r_[self.phi, other.phi]
        theta = np.r_[self.theta, other.theta]

        return Plan(time=time, phi=phi, theta=theta, frame=self.frame.name, site=self.site)

    def __radd__(self, other):
        return other.__add__(self)



from __future__ import annotations

import matplotlib.pyplot as plt
import pytest

import maria

plt.close("all")

# subarray_dets = {
#     "subarray-1": {
#         "n": 500,
#         "field_of_view": 2,
#         "array_shape": "hex",
#         "bands": [{"center": 30, "width": 5}, {"center": 40, "width": 5}],
#     },
#     "subarray-2": {
#         "n": 500,
#         "field_of_view": 2,
#         "array_shape": "hex",
#         "bands": [{"center": 90, "width": 5}, {"center": 150, "width": 5}],
#     },
# }

# predefined_bands = {
#     "n": 500,
#     "field_of_view": 2,
#     "array_packing": "square",
#     "array_shape": "hex",
#     "bands": ["alma/f043", "alma/f078"],
# }

# predefined_dets = {"file": "data/alma/alma.cycle1.total.csv"}

# beam_packing_dets = {
#     "subarray-1": {
#         "field_of_view": 0.2,
#         "beam_spacing": 1,
#         "primary_size": 10,
#         "array_shape": "hex",
#         "array_packing": "sunflower",
#         "band": {"center": 30, "width": 5},
#     },
#     "subarray-2": {
#         "n": 500,
#         "field_of_view": 0.2,
#         "array_shape": "circle",
#         "array_packing": "hex",
#         "bands": [{"center": 90, "width": 5}, {"center": 150, "width": 5}],
#     },
# }


@pytest.mark.parametrize("instrument_name", maria.all_instruments)
def test_get_instrument(instrument_name):
    instrument = maria.get_instrument(instrument_name)
    print(instrument)

    instrument.plot()


# @pytest.mark.parametrize(
#     "dets", [subarray_dets, predefined_bands, predefined_dets, beam_packing_dets]
# )
# def test_get_custom_array(dets):
#     instrument = maria.get_instrument(dets=dets)



from __future__ import annotations

import logging

import numpy as np

logger = logging.getLogger("maria")

UNITS = {
    "radians": {"short_name": "rad", "factor": 1.0, "symbol": "rad"},
    "degrees": {"short_name": "deg", "factor": 180 / np.pi, "symbol": "°"},
    "arcminutes": {"short_name": "arcmin", "factor": 60 * 180 / np.pi, "symbol": "’"},
    "arcseconds": {"short_name": "arcsec", "factor": 3600 * 180 / np.pi, "symbol": "”"},
}


class Angle:
    def __init__(self, a, units: str = "radians", unwrap: bool = False):
        self.x = None
        self.unwrap = unwrap
        for k in UNITS:
            if units in [k, UNITS[k]["short_name"]]:
                self.x = np.array(a) / UNITS[k]["factor"]
        if self.x is None:
            raise ValueError(f"Invalid units '{units}'.")

        self.is_scalar = len(np.shape(self.radians)) == 0

        if self.is_scalar and unwrap:
            raise ValueError()

        # 3 April 2025
        logger.warning(
            "The Angle class is deprecated and will be removed in a future update. "
            "Please refactor your code using the Quantity class (maria.units.Quantity)."
        )

    def __getattr__(self, attr):
        radians = self.x if self.unwrap else (self.x + np.pi) % (2 * np.pi) - np.pi
        for k in UNITS:
            if attr in [k, UNITS[k]["short_name"]]:
                return radians * UNITS[k]["factor"]
        raise AttributeError(f"'Angle' object has no attribute '{attr}'.")

    def __getitem__(self, idx):
        return self.radians[idx]

    def __float__(self):
        return self.radians

    def __repr__(self):
        units = self.units
        if self.is_scalar:
            return f"{round(getattr(self, units), 3)}{self.symbol}"
        else:
            return f"Angle({getattr(self, units)}, units={units})"

    @property
    def shape(self):
        return self.radians.shape

    @property
    def symbol(self):
        return UNITS[self.units]["symbol"]

    @property
    def units(self):
        # peak-to-peak
        max_deg = self.deg if self.is_scalar else self.deg.max()
        if max_deg < 0.5 / 60:
            return "arcseconds"
        if max_deg < 0.5:
            return "arcminutes"
        return "degrees"

    @property
    def units_short(self):
        return UNITS[self.units]["short_name"]

    @property
    def values(self):
        return getattr(self, self.units)



from __future__ import annotations

import functools
import logging
import os
import time as ttime
from copy import deepcopy

import arrow
import dask.array as da
import numpy as np
import pandas as pd
import scipy as sp
from astropy import units as u
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.time import Time
from scipy.interpolate import interp1d

from ..io import DEFAULT_TIME_FORMAT, humanize_time, repr_lat_lon
from ..units import Quantity
from ..utils import get_utc_day_hour, get_utc_year_day
from .frame import FRAMES, Frame
from .transforms import (
    get_center_phi_theta,
    offsets_to_phi_theta,
    phi_theta_to_offsets,
    phi_theta_to_xyz,
    unjitted_offsets_to_phi_theta,
    xyz_to_phi_theta,
)

logger = logging.getLogger("maria")
here, this_filename = os.path.split(__file__)
frames = pd.read_csv(f"{here}/frames.csv", index_col=0).T

DEFAULT_EARTH_LOCATION = EarthLocation.from_geodetic(
    0.0,
    90.0,
    height=0.0,
    ellipsoid=None,
)  # noqa
DEFAULT_TIMESTAMP = arrow.now().timestamp()


class Coordinates:
    """
    A class for managing coordinates, allowing us to access different coordinate frames at once.
    """

    def __init__(
        self,
        phi: float = 0.0,
        theta: float = 0.0,
        r: float = 0.0,
        t: float = DEFAULT_TIMESTAMP,
        x: float = 0.0,
        y: float = 0.0,
        z: float = 0.0,
        earth_location: EarthLocation = DEFAULT_EARTH_LOCATION,
        frame: str = "az/el",
        dtype: type = np.float64,
    ):
        self.earth_location = earth_location
        self.frame = Frame(frame)
        self.dtype = dtype

        if isinstance(t, str):
            t = arrow.get(t).timestamp()

        # DO NOT BROADCAST TIME. IT STAYS ONE-DIMENSIONAL.
        for attr, value in zip(
            ["x", "y", "z", "r", "phi", "theta", "t"],
            np.broadcast_arrays(x, y, z, r, phi, theta, t),
        ):
            # if not isinstance(value, dask.array.Array):
            #     value = da.asarray(value)
            # setattr(self, f"_{attr}", da.asarray(value))
            setattr(self, f"_{attr}", value)

        setattr(self, self.frame.phi_name, self._phi)
        setattr(self, self.frame.theta_name, self._theta)

        if hasattr(t, "__len__"):
            for axis in range(len(self.t.shape) - 1):
                if (np.ptp(self.t, axis=axis) > 0).any():
                    raise ValueError("Only the last axis can vary in time.")

        self.initialized = False
        self.computed_frames = [self.frame.name]

    def initialize(self):
        """
        Do all the initial computations needed to convert to some other frame.
        """

        ref_time = ttime.monotonic()

        self.shaped_t = np.atleast_1d(self.t)
        keep_dims = (-1,) if hasattr(self.t, "__len__") else ()
        t_ordered_center_phi_theta = np.c_[get_center_phi_theta(self._phi, self._theta, keep_dims=keep_dims)]

        # (nt) t samples on which to explicitly compute the transformation from astropy
        t_samples_min_res_seconds = 60
        t_samples_min = np.min(self.t) - 1e0
        t_samples_max = np.max(self.t) + 1e0
        n_t_samples = int(
            np.maximum(
                3,
                (t_samples_max - t_samples_min) / t_samples_min_res_seconds,
            ),
        )

        n_t_samples = np.minimum(n_t_samples, 10000)

        if 3 <= self.t.size < n_t_samples:
            self.fid_t = self.t
        else:
            self.fid_t = np.linspace(t_samples_min, t_samples_max, n_t_samples)

        sample_indices = interp1d(
            self.shaped_t,
            np.arange(len(self.shaped_t)),
            bounds_error=False,
            kind="nearest",
            fill_value="extrapolate",
        )(self.fid_t).astype(int)

        # three fiducial offsets from the boresight to train a transformation matrix
        # shape: (n_fid, 2)
        fid_offsets = np.radians([[-1.0, 0.0], [1.0, 0.0], [0.0, 1.0]])
        psi = np.linspace(0, 2 * np.pi, 12 + 1)[:-1]
        fid_offsets = np.concatenate(
            [r * np.c_[np.cos(psi), np.sin(psi)] for r in [1, 10, 30]],
            axis=0,
        )

        # for sample_index =

        fpt = unjitted_offsets_to_phi_theta(
            fid_offsets[..., :2],
            t_ordered_center_phi_theta[:, 0][sample_indices][..., None],
            t_ordered_center_phi_theta[:, 1][sample_indices][..., None],
        )

        self.fid_phi, self.fid_theta = fpt[..., 0], fpt[..., 1]

        # self.fid_phi = np.zeros((n_t_samples, len(fid_offsets)))
        # self.fid_theta = np.zeros((n_t_samples, len(fid_offsets)))

        # for ii, it in enumerate(sample_indices):
        #     self.fid_phi[ii], self.fid_theta[ii] = offsets_to_phi_theta(
        #     fid_offsets[..., :1],
        #     *t_ordered_center_phi_theta[it],
        # )

        self.fid_skycoords = {
            self.frame.name: SkyCoord(
                self.fid_phi * u.rad,
                self.fid_theta * u.rad,
                obstime=Time(
                    self.fid_t[:, None],
                    format="unix",
                ),
                frame=frames[self.frame.name]["astropy_name"],
                location=self.earth_location,
            ),
        }

        self.transforms = {}
        self.fid_points = {self.frame.name: phi_theta_to_xyz(self.fid_phi, self.fid_theta)}
        self.A = self.fid_points[self.frame.name]
        self.AT = np.swapaxes(self.fid_points[self.frame.name], -2, -1)

        self.initialized = True

        duration_s = ttime.monotonic() - ref_time
        logger.debug(
            f"Initialized {self} in {humanize_time(duration_s)}.",
        )  # noqa

    def compute_transform(self, frame):
        compute_transform_start_s = ttime.monotonic()

        frame = Frame(frame)

        if not self.initialized:
            self.initialize()

        self.fid_skycoords[frame.name] = getattr(
            self.fid_skycoords[self.frame.name],
            frame.astropy_name,
        )

        frame_fid_phi = getattr(
            self.fid_skycoords[frame.name],
            frame.astropy_phi,
        ).rad
        frame_fid_theta = getattr(
            self.fid_skycoords[frame.name],
            frame.astropy_theta,
        ).rad

        self.fid_points[frame.name] = phi_theta_to_xyz(frame_fid_phi, frame_fid_theta)

        # voodoo!
        self.transforms[frame.name] = (
            np.linalg.inv(self.AT @ self.A) @ self.AT @ phi_theta_to_xyz(frame_fid_phi, frame_fid_theta)
        )

        transform_stack = interp1d(
            self.fid_t,
            self.transforms[frame.name],
            kind="quadratic",
            bounds_error=False,
            fill_value="extrapolate",
            axis=0,
        )(self.t)

        frame_phi, frame_theta = xyz_to_phi_theta(
            (np.expand_dims(self.compute_points(), -2) @ transform_stack).squeeze(),
        )

        setattr(self, frame.phi_name, frame_phi)
        setattr(self, frame.theta_name, frame_theta)

        self.computed_frames.append(frame.name)

        duration = ttime.monotonic() - compute_transform_start_s
        logger.debug(
            f"Computed transform to frame '{frame.name}' for {self} in {humanize_time(duration)}.",
        )  # noqa

    @property
    def shape(self):
        return self._phi.shape

    @property
    def size(self):
        return self._phi.size

    @property
    def ndim(self):
        return self._phi.ndim

    def __getattr__(self, attr):
        if attr == "t":
            return self._t[tuple(0 for _ in range(self.ndim - 1))]

        if attr in ["x", "y", "z", "r", "phi", "theta"]:
            return getattr(self, f"_{attr}")

        for frame_name, config in FRAMES.items():
            if attr in [config["phi"]["name"], config["theta"]["name"]]:
                self.compute_transform(frame=frame_name)
                return getattr(self, attr)

        raise AttributeError(f"'Coordinates' object has no attribute '{attr}'")

    def __getitem__(self, key):
        clone = deepcopy(self)
        attrs = ["_x", "_y", "_z", "_r", "_phi", "_theta", "_t"]
        attrs.extend([frames[frame][c] for frame in self.computed_frames for c in ["phi", "theta"]])
        for attr in attrs:
            setattr(clone, attr, getattr(clone, attr)[key])

        return clone

    @property
    def day_hour(self):
        return np.array(list(map(get_utc_day_hour, self.t)))

    @property
    def year_day(self):
        return np.array(list(map(get_utc_year_day, self.t)))

    @property
    def timestep(self):
        if len(self.t):
            return np.mean(np.gradient(self.t))
        return None

    def downsample(self, timestep: float = None, factor: int = None):
        if timestep is None and factor is None:
            raise ValueError("You must supply either 'timestep' or 'factor'.")

        timestep = timestep or factor * self.timestep

        ds_t = np.arange(self.t.min(), self.t.max(), timestep)
        ds_phi = sp.interpolate.interp1d(self.t, self._phi, axis=-1, bounds_error=False, fill_value="extrapolate")(ds_t)

        ds_theta = sp.interpolate.interp1d(self.t, self._theta, axis=-1, bounds_error=False, fill_value="extrapolate")(ds_t)

        return Coordinates(
            t=ds_t,
            phi=ds_phi,
            theta=ds_theta,
            earth_location=self.earth_location,
            frame=self.frame.name,
            dtype=self.dtype,
        )

    def boresight(self):
        cphi, ctheta = get_center_phi_theta(self._phi, self.theta, keep_dims=(-1,))

        return Coordinates(
            t=self.t,
            phi=cphi,
            theta=ctheta,
            earth_location=self.earth_location,
            frame=self.frame.name,
            dtype=self.dtype,
        )

    @functools.cached_property
    def summary(self):
        # compute summary for the string repr
        summary = pd.DataFrame(columns=["min", "mean", "max"])
        boresight = self.boresight
        for attr in ["az", "el", "ra", "dec"]:
            for stat in ["min", "mean", "max"]:
                summary.loc[attr, stat] = f"{float(np.degrees(getattr(getattr(boresight, attr), stat)())):.03f}°"

        return summary

    @property
    def xyz(self):
        return np.concatenate([self.x[..., None], self.y[..., None], self.z[..., None]], axis=-1)

    def project(self, z, frame="az/el"):
        frame = Frame(frame)

        phi = getattr(self, frame.phi_name)
        theta = getattr(self, frame.theta_name)

        tan_theta = np.tan(theta)[..., None]
        p = (z - self.z)[..., None] * np.concatenate(
            [
                np.cos(phi)[..., None] / tan_theta,
                np.sin(phi)[..., None] / tan_theta,
                np.ones((*phi.shape, 1)),
            ],
            axis=-1,
        )

        return p + self.xyz

    def compute_points(self):
        return phi_theta_to_xyz(self._phi, self._theta)

    def center(self, frame=None):
        frame = Frame(frame or self.frame)
        return np.array(
            get_center_phi_theta(
                getattr(self, frame.phi_name),
                getattr(self, frame.theta_name),
            )
        )

    def broadcast(self, offsets, frame="az/el"):
        phi_theta = unjitted_offsets_to_phi_theta(offsets[..., None, :], self.az, self.el)
        return Coordinates(
            t=self.t,
            phi=phi_theta[..., 0],
            theta=phi_theta[..., 1],
            earth_location=self.earth_location,
            frame="az/el",
        )

    def offsets(self, frame, center="auto", units="radians", compute: bool = False):
        offsets_s = ttime.monotonic()

        frame = Frame(frame)

        if compute:
            return self.offsets(frame=frame, center=center, units=units, compute=False)

        if isinstance(center, str):
            if center == "auto":
                center = self.center(frame=frame)
        if frame.name == "az/el":
            X = phi_theta_to_offsets(np.stack([self.az, self.el], axis=-1), *center)
        elif frame.name == "ra/dec":
            X = phi_theta_to_offsets(np.stack([self.ra, self.dec], axis=-1), *center)
        elif frame.name == "galactic":
            X = phi_theta_to_offsets(np.stack([self.l, self.b], axis=-1), *center)

        logger.debug(f"Computed offsets for {self} in {humanize_time(ttime.monotonic() - offsets_s)}")

        return X

    def spread(self, frame="ra/dec"):
        dX = self.offsets(frame=frame)
        return dX.std(axis=list(range(1, dX.ndim)))

    def __repr__(self):
        lon = self.earth_location.lon.deg
        lat = self.earth_location.lat.deg

        date_string = arrow.get(np.mean(self.t)).to("utc").format(DEFAULT_TIME_FORMAT)

        return f"Coordinates(shape={self.shape}, earth_location={repr_lat_lon(lat, lon)}, time='{date_string} UTC')"

    @property
    def longitude(self):
        return Quantity(self.earth_location.lon.deg, "deg")

    @property
    def latitude(self):
        return Quantity(self.earth_location.lat.deg, "deg")

    def __repr__(self):
        repr_lat, repr_lon = repr_lat_lon(self.latitude.degrees, self.longitude.degrees)
        return f"""Coordinates:
  shape: {self.shape}
  location:
    lat: {repr_lat}
    lon: {repr_lon}
  time:
    min: {arrow.get(self.t.min()).format(DEFAULT_TIME_FORMAT)}
    max: {arrow.get(self.t.max()).format(DEFAULT_TIME_FORMAT)}"""



from __future__ import annotations

import logging
import time as ttime

import numpy as np
import scipy as sp

from .. import utils
from ..io import humanize_time
from ..utils import remove_slope
from .tod import TOD

logger = logging.getLogger("maria")

OPERATION_KWARGS = {
    "remove_slope": {},
    "window": {
        "name": {"dtype": str, "aliases": ["window"]},
        "kwargs": {"dtype": dict, "aliases": ["window_kwargs"]},
    },
    "filter": {
        "f_lower": {"dtype": float, "aliases": ["f_lower"]},
        "f_upper": {"dtype": float, "aliases": ["f_upper"]},
        "order": {"dtype": int, "aliases": ["filter_order"]},
        "method": {"dtype": str, "aliases": ["filter_method"]},
    },
    "remove_modes": {
        "modes_to_remove": {"dtype": list, "aliases": ["modes_to_remove"]},
    },
    "remove_spline": {
        "knot_spacing": {"dtype": float, "aliases": ["remove_spline_knot_spacing"]},
        "remove_el_gradient": {"dtype": bool, "aliases": ["remove_el_gradient"]},
        "order": {"dtype": int, "aliases": ["depline_order"]},
    },
}


def process_operation_kwargs(**kwargs):  # lol
    config = {}

    for operation, operation_params in OPERATION_KWARGS.items():
        if operation not in OPERATION_KWARGS:
            raise ValueError(f'Invalid operation "{operation}". Valid operations are {OPERATION_KWARGS.keys()}.')

        subconfig = {}
        for key, param in operation_params.items():
            for kwarg in list(kwargs.keys()):
                if kwarg in param["aliases"]:
                    subconfig[key] = kwargs.pop(kwarg)
                    continue

        if subconfig:
            config[operation] = subconfig

    if len(kwargs) > 0:
        raise ValueError(f"Invalid kwargs for TOD processing: {kwargs}.")

    return config


def validate_process_config(config):
    for operation, operation_params in config.items():
        if operation not in OPERATION_KWARGS:
            raise ValueError(
                f"Invalid operation '{operation}'. Valid operations are {list(OPERATION_KWARGS.keys())}",
            )

        for key, value in operation_params.items():
            if key not in OPERATION_KWARGS[operation]:
                raise ValueError(
                    f"Invalid param '{key}' for operation '{operation}'. "
                    f"Valid parameters for this operation are {list(OPERATION_KWARGS[operation].keys())}",
                )

            dtype = OPERATION_KWARGS[operation][key]["dtype"]

            if not isinstance(value, dtype):
                try:
                    config[operation][key] = dtype(value)
                except Exception:
                    param = {key: value}
                    raise TypeError(
                        f"Could not convert param {param} for operation '{operation}' to requisite type '{dtype.__name__}'.",
                    )

    return config


def process_tod(tod, config=None, **kwargs):
    config = validate_process_config(config or process_operation_kwargs(**kwargs))

    D = tod.signal.compute()
    W = np.ones(D.shape)

    if "remove_slope" in config:
        remove_slope_start_s = ttime.monotonic()
        D -= np.linspace(D[..., 0], D[..., -1], D.shape[-1]).T

        logger.debug(f'Completed tod operation "remove_slope in {humanize_time(ttime.monotonic() - remove_slope_start_s)}.')
        if np.isnan(D).any():
            raise ValueError("tod operation 'remove_slope' introduced NaNs")

    if "remove_spline" in config:
        remove_spline_start_s = ttime.monotonic()

        B = utils.signal.bspline_basis(
            tod.time,
            spacing=config["remove_spline"]["knot_spacing"],
            order=config["remove_spline"].get("order", 3),
        )

        if config["remove_spline"].get("remove_el_gradient", False):
            el = tod.boresight.el
            el_ptp = el.max() - el.min()
            if el_ptp == 0:
                raise ValueError("Cannot remove elevation gradient when elevation is constant")
            rel_el = (el - el.min()) / el_ptp
            B = np.concatenate([B * rel_el[None], B * (1 - rel_el)[None]], axis=0)

        A = np.linalg.inv(B @ B.T) @ B @ D.T
        D -= A.T @ B

        logger.debug(
            f'Completed tod operation "remove_spline" in {humanize_time(ttime.monotonic() - remove_spline_start_s)}.'
        )
        if np.isnan(D).any():
            raise ValueError("tod operation 'remove_spline' introduced NaNs")

    if "window" in config:
        window_start_s = ttime.monotonic()
        window_function = getattr(sp.signal.windows, config["window"]["name"])
        W *= window_function(D.shape[-1], **config["window"].get("kwargs", {}))
        D *= W
        logger.debug(f'Completed tod operation "window" in {humanize_time(ttime.monotonic() - window_start_s)}.')
        if np.isnan(D).any():
            raise ValueError("tod operation 'window' introduced NaNs")

    if "filter" in config:
        filter_start_s = ttime.monotonic()

        D = remove_slope(D)

        if "window" not in config:
            logger.warning("Filtering without windowing is not recommended.")

        if "f_upper" in config["filter"]:
            D = utils.signal.lowpass(
                D,
                fc=config["filter"]["f_upper"],
                sample_rate=tod.sample_rate,
                order=config["filter"].get("order", 1),
                method="bessel",
            )

        if "f_lower" in config["filter"]:
            D = utils.signal.highpass(
                D,
                fc=config["filter"]["f_lower"],
                sample_rate=tod.sample_rate,
                order=config["filter"].get("order", 1),
                method="bessel",
            )

        logger.debug(f'Completed tod operation "filter" in {humanize_time(ttime.monotonic() - filter_start_s)}.')
        if np.isnan(D).any():
            raise ValueError("tod operation 'filter' introduced NaNs")

    if "remove_modes" in config:
        remove_modes_start_s = ttime.monotonic()

        modes_to_remove = config["remove_modes"]["modes_to_remove"]
        A, B = utils.signal.decompose(D, k=np.max(modes_to_remove) + 1)
        D -= A[:, modes_to_remove] @ B[modes_to_remove]

        logger.debug(f'Completed tod operation "remove_modes" in {humanize_time(ttime.monotonic() - remove_modes_start_s)}.')
        if np.isnan(D).any():
            raise ValueError("tod operation 'remove_modes' introduced NaNs")

    ptod = TOD(
        data={"total": D},
        weight=W,
        coords=tod.coords,
        units=tod.units,
        dets=tod.dets,
        dtype=np.float32,
        metadata=tod.metadata,
    )

    ptod.processing_config = config

    return ptod



import logging
import os

from ..units import Quantity

logger = logging.getLogger("maria")
here, this_filename = os.path.split(__file__)


def repr_lat_lon(lat, lon):
    lat_repr = Quantity(abs(lat), "deg").dms + (" N" if lat > 0 else " S")
    lon_repr = Quantity(abs(lon), "deg").dms + (" E" if lon > 0 else " W")
    return lat_repr, lon_repr


def repr_phi_theta(phi, theta, frame, join=None):
    qphi = Quantity(phi, "rad")
    qtheta = Quantity(theta, "rad")
    if frame == "az/el":
        res = (f"az: {qphi.deg:.02f}°", f"el: {qtheta.deg:.02f}°")
    elif frame == "ra/dec":
        res = (f"ra:  {qphi.hms}", f"dec: {qtheta.dms}")
    elif frame == "galactic":
        res = (f"l: {qphi.deg:.02f}°", f"b: {qtheta.deg:.02f}°")
    else:
        raise ValueError(f"Invalid frame '{frame}'")

    if join is not None:
        res = join.join(res)

    return res



import arrow


def get_day_hour(t):
    a = arrow.get(t)
    return a.hour + a.minute / 60 + a.second / 3600 + int(a.format("SSS")) / 3600000


def get_utc_day_hour(t):
    return get_day_hour(arrow.get(t).to("utc"))


def get_utc_year_day(t, partial=True):
    whole_day = int(arrow.get(t).to("utc").format("DDD")) - 1
    if partial:
        return whole_day + get_utc_day_hour(t) / 24
    return whole_day


def get_utc_year(t):
    return arrow.get(t).to("utc").year



from __future__ import annotations

import logging
import os
import time as ttime

import arrow
import numpy as np
from tqdm import tqdm

from ..atmosphere import DEFAULT_ATMOSPHERE_KWARGS, Atmosphere
from ..cmb import CMB, DEFAULT_CMB_KWARGS, generate_cmb, get_cmb
from ..coords import Coordinates
from ..errors import PointingError
from ..instrument import Instrument, get_instrument
from ..io import fetch, humanize_time
from ..map import Map, load
from ..plan import Plan, PlanList, get_plan
from ..site import Site, get_site
from ..tod import TOD
from ..utils import get_rotation_matrix_2d, read_yaml

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")

MIN_ELEVATION_WARN = 10  # degrees
MIN_ELEVATION_ERROR = 5  # degrees


class Observation:
    """
    A simulation of a telescope. This is what users should touch, primarily.
    """

    def __init__(
        self,
        instrument: Instrument,
        plan: Plan,
        site: Site,
        atmosphere: str = None,
        atmosphere_kwargs: dict = {},
        weather_kwargs: dict = {},
    ):
        self.instrument = instrument
        self.plan = plan
        self.site = site

        self.weather_kwargs = weather_kwargs

        obs_init_s = ttime.monotonic()

        self.boresight = Coordinates(
            t=self.plan.time,
            phi=self.plan.phi,
            theta=self.plan.theta,
            earth_location=self.site.earth_location,
            frame=self.plan.frame,
        )

        # this can be expensive sometimes
        self.coords = self.boresight.broadcast(
            self.instrument.dets.offsets @ get_rotation_matrix_2d(np.radians(plan.roll)).T,
            frame="az/el",
        )

        # do some checks
        el_min = np.atleast_1d(self.coords.el).min()
        if el_min < np.radians(MIN_ELEVATION_WARN):
            logger.warning(
                f"Some detectors come within {MIN_ELEVATION_WARN} degrees of the horizon"
                f"(el_min = {np.degrees(el_min):.01f}°)",
            )
        if el_min <= np.radians(MIN_ELEVATION_ERROR):
            raise PointingError(
                f"Some detectors come within {MIN_ELEVATION_ERROR} degrees of the horizon"
                f"(el_min = {np.degrees(el_min):.01f}°)",
            )

        max_abs_az_vel_deg = np.max(
            np.abs(np.gradient(np.degrees(np.unwrap(self.boresight.az))) / np.gradient(self.boresight.t))
        )
        max_abs_el_vel_deg = np.max(np.abs(np.gradient(np.degrees(self.boresight.el)) / np.gradient(self.boresight.t)))

        if max_abs_az_vel_deg > self.instrument.az_vel_limit:
            logger.warning(
                (
                    f"The maximum velocity of the boresight ({max_abs_az_vel_deg:.01f} deg/s) exceeds "
                    f"the maximum velocity of the instrument ({self.instrument.az_vel_limit:.01f} deg/s)."
                ),
            )

        logger.debug(f"Initialized observation coordinates in {humanize_time(ttime.monotonic() - obs_init_s)}.")

        if atmosphere:
            self.atmosphere_kwargs = atmosphere_kwargs
            self.atmosphere = Atmosphere(
                model=atmosphere,
                timestamp=self.plan.time.mean(),
                region=self.site.region,
                altitude=self.site.altitude,
                weather_kwargs=self.weather_kwargs,
                **self.atmosphere_kwargs,
            )

        self.loading = {}

    @property
    def shape(self):
        return self.coords.shape


#         super().__init__(
#             instrument=instrument,
#             plan=plan,
#             site=site,
#             progress_bars=progress_bars,
#             keep_mean_signal=keep_mean_signal,
#         )

#         sim_start_s = ttime.monotonic()

#         self.noise = noise

#         self.map_kwargs = map_kwargs
#         self.noise_kwargs = noise_kwargs

#         self.start = arrow.get(self.boresight.t.min()).to("utc")
#         self.end = arrow.get(self.boresight.t.max()).to("utc")

#         cmb_start_s = ttime.monotonic()

#         if cmb:
#             self.cmb_kwargs = DEFAULT_CMB_KWARGS.copy()
#             self.cmb_kwargs.update(cmb_kwargs)

#             if cmb in ["spectrum", "power_spectrum", "generate", "generated"]:
#                 for _ in tqdm(
#                     range(1),
#                     desc=f"Generating CMB (nside={self.cmb_kwargs['nside']})",
#                     disable=self.disable_progress_bars,
#                 ):
#                     self.cmb = generate_cmb(**self.cmb_kwargs)
#             elif cmb in ["real", "planck"]:
#                 self.cmb = get_cmb(**self.cmb_kwargs)
#             else:
#                 raise ValueError(f"Invalid value for cmb: '{cmb}'.")

#             logger.debug(f"Initialized CMB simulation in {humanize_time(ttime.monotonic() - cmb_start_s)}.")

#         map_start_s = ttime.monotonic()
#         if map:
#             if isinstance(map, str):
#                 map = load(fetch(map), **map_kwargs)
#             elif not isinstance(map, Map):
#                 raise ValueError("")

#             # the map can be frequency-naive if it is already in K_RJ
#             self.map = map.to(units="K_RJ")

#             if "stokes" not in self.map.dims:
#                 self.map = self.map.unsqueeze("stokes")

#             if "nu" not in self.map.dims:
#                 self.map = self.map.unsqueeze("nu")

#             if "t" in self.map.dims:
#                 if self.map.dims["t"] > 1:
#                     map_start = arrow.get(self.map.t.min()).to("utc")
#                     map_end = arrow.get(self.map.t.max()).to("utc")
#                     if map_start > self.start:
#                         logger.warning(
#                             f"Beginning of map ({map_start.isoformat()[:26]}) is after the "
#                             f"beginning of the simulation ({self.start.isoformat()[:26]}).",
#                         )
#                     if map_end < self.end:
#                         logger.warning(
#                             f"End of map ({map_end.isoformat()[:26]}) is before the "
#                             f"end of the simulation ({self.end.isoformat()[:26]}).",
#                         )
#                 else:
#                     self.map = self.map.squeeze("t")

#         logger.debug(f"Initialized map simulation in {humanize_time(ttime.monotonic() - map_start_s)}.")

#         noise_start_s = ttime.monotonic()
#         if noise:
#             pass
#         logger.debug(f"Initialized noise simulation in {humanize_time(ttime.monotonic() - noise_start_s)}.")

#         if atmosphere:
#             atmosphere_init_start_s = ttime.monotonic()

#             self.atmosphere_kwargs = DEFAULT_ATMOSPHERE_KWARGS.copy()
#             self.atmosphere_kwargs.update(atmosphere_kwargs)

#             # do some checks
#             el_min = np.atleast_1d(self.coords.el).min()
#             if el_min < np.radians(MIN_ELEVATION_WARN):
#                 logger.warning(
#                     f"Some ArrayList come within {MIN_ELEVATION_WARN} degrees of the horizon"
#                     f"(el_min = {np.degrees(el_min):.01f}°)",
#                 )
#             if el_min <= np.radians(MIN_ELEVATION_ERROR):
#                 raise PointingError(
#                     f"Some ArrayList come within {MIN_ELEVATION_ERROR} degrees of the horizon"
#                     f"(el_min = {np.degrees(el_min):.01f}°)",
#                 )

#             self.weather_kwargs = self.atmosphere_kwargs.pop("weather") if "weather" in self.atmosphere_kwargs else {}

#             self.atmosphere = Atmosphere(
#                 model=atmosphere,
#                 timestamp=self.plan.time.mean(),
#                 region=self.site.region,
#                 altitude=self.site.altitude,
#                 weather_kwargs=self.weather_kwargs,
#                 disable_progress_bars=self.disable_progress_bars,
#                 **self.atmosphere_kwargs,
#             )

#             # give it the simulation, so that it knows about pointing, site, etc. (kind of cursed)
#             self.atmosphere.initialize(self)

#             logger.debug(
#                 f"Initialized atmosphere simulation in {humanize_time(ttime.monotonic() - atmosphere_init_start_s)}."
#             )

#         logger.debug(f"Initialized simulation in {humanize_time(ttime.monotonic() - sim_start_s)}.")

#     def _run(self):
#         if hasattr(self, "atmosphere"):
#             if not self.atmosphere._initialized:
#                 atmosphere_init_start_s = ttime.monotonic()
#                 self.atmosphere.initialize(self)
#                 logger.debug(
#                     f"Initialized atmosphere simulation in {humanize_time(ttime.monotonic() - atmosphere_init_start_s)}."
#                 )

#             atmosphere_sim_start_s = ttime.monotonic()
#             self._simulate_atmosphere()
#             self._compute_atmospheric_loading()
#             logger.debug(f"Ran atmosphere simulation in {humanize_time(ttime.monotonic() - atmosphere_sim_start_s)}.")

#         if hasattr(self, "cmb"):
#             cmb_sim_start_s = ttime.monotonic()
#             self._compute_cmb_loading()
#             logger.debug(f"Ran CMB simulation in {humanize_time(ttime.monotonic() - cmb_sim_start_s)}.")

#         if hasattr(self, "map"):
#             map_sim_start_s = ttime.monotonic()
#             self._sample_maps()
#             logger.debug(f"Ran map simulation in {humanize_time(ttime.monotonic() - map_sim_start_s)}.")

#         # number of bands are lost here
#         if self.noise:
#             noise_sim_start_s = ttime.monotonic()
#             self._simulate_noise()
#             logger.debug(f"Ran noise simulation in {humanize_time(ttime.monotonic() - noise_sim_start_s)}.")

#         gain_error = np.exp(
#             self.instrument.dets.gain_error * np.random.standard_normal(size=self.instrument.dets.n),
#         )

#         for field in self.loading:
#             if field in ["noise"]:
#                 continue

#             self.loading[field] *= gain_error[:, None]

#     def __repr__(self):
#         instrument_tree = "├ " + self.instrument.__repr__().replace("\n", "\n│ ")
#         site_tree = "├ " + self.site.__repr__().replace("\n", "\n│ ")
#         plan_tree = "├ " + self.plan.__repr__().replace("\n", "\n│ ")

#         trees = []
#         attrs = [attr for attr in ["atmosphere", "cmb", "map"] if getattr(self, attr, None)]
#         if attrs:
#             for attr in attrs[:-1]:
#                 trees.append("├ " + getattr(self, attr).__repr__().replace("\n", "\n│ "))
#             trees.append("└ " + getattr(self, attrs[-1]).__repr__().replace("\n", "\n  "))

#         trees_string = "\n".join(trees)

#         return f"""Simulation
# {instrument_tree}
# {site_tree}
# {plan_tree}
# {trees_string}"""

#     @property
#     def total_loading(self):
#         return sum([d for d in self.loading.values()])



from __future__ import annotations

import os

import arrow
import h5py
import numpy as np
import pandas as pd
import scipy as sp

from ..constants import g
from ..io import DEFAULT_TIME_FORMAT, fetch
from ..site import REGIONS, InvalidRegionError, all_regions
from ..units import Quantity
from ..utils import get_utc_day_hour, get_utc_year_day

here, this_filename = os.path.split(__file__)

WEATHER_CACHE_BASE = "/tmp/maria-data/weather"
WEATHER_SOURCE_BASE = "https://github.com/thomaswmorris/maria-data/raw/master/atmosphere/weather"  # noqa F401


def get_vapor_pressure(air_temp, rel_hum):  # units are (°K, %)
    T = air_temp - 273.15  # in °C
    a, b, c = 611.21, 17.67, 238.88  # units are Pa, ., °C
    gamma = np.log(1e-2 * rel_hum) + b * T / (c + T)
    return a * np.exp(gamma)


def get_saturation_pressure(air_temp):  # units are (°K, %)
    T = air_temp - 273.15  # in °C
    a, b, c = 611.21, 17.67, 238.88  # units are Pa, ., °C
    return a * np.exp(b * T / (c + T))


def get_dew_point(air_temp, rel_hum):  # units are (°K, %)
    a, b, c = 611.21, 17.67, 238.88  # units are Pa, ., °C
    p_vap = get_vapor_pressure(air_temp, rel_hum)
    return c * np.log(p_vap / a) / (b - np.log(p_vap / a)) + 273.15


def get_relative_humidity(air_temp, dew_point):
    T, DP = air_temp - 273.15, dew_point - 273.15  # in °C
    b, c = 17.67, 238.88
    return 1e2 * np.exp(b * DP / (c + DP) - b * T / (c + T))


def relative_to_absolute_humidity(air_temp, rel_hum):
    return 1e-2 * rel_hum * get_saturation_pressure(air_temp) / (461.5 * air_temp)


def absolute_to_relative_humidity(air_temp, abs_hum):
    return 1e2 * 461.5 * air_temp * abs_hum / get_saturation_pressure(air_temp)


class Weather:
    def __init__(
        self,
        region: str = "chajnantor",
        time: arrow.Arrow = None,
        altitude: float = None,
        quantiles: dict = {},
        override: dict = {},
        source: str = "era5",
        refresh_cache: bool = False,
    ):
        self.data = {}

        if region not in all_regions:
            raise InvalidRegionError(region)

        if altitude is None:
            altitude = REGIONS.loc[region, "altitude"]

        self.region = region
        self.base_altitude = Quantity(altitude, "m").m
        self.quantiles = quantiles
        self.override = override
        self.source = source

        time = time if time is not None else arrow.now().to("utc")
        self.time = arrow.get(time)

        self.cache_path = fetch(
            f"atmosphere/weather/{source}/{self.region}.h5",
            max_age=30 * 86400,
            refresh=refresh_cache,
        )

        self.time_zone = REGIONS.loc[self.region, "timezone"]
        self.local_time = self.time.to(self.time_zone)

        self.utc_day_hour = get_utc_day_hour(self.time.timestamp())
        self.utc_year_day = get_utc_year_day(self.time.timestamp())

        self.region_quantiles = {}

        with h5py.File(self.cache_path, "r") as f:
            self.quantile_levels = f["quantile_levels"][:]
            self.pressure_levels = f["pressure_levels"][:]

            self.fields = list(f["data"].keys())

            # for fun sampling things
            year_day_side = f["year_day_side"][:]
            day_hour_side = f["day_hour_side"][:]
            year_day_edge_index = f["year_day_edge_index"][:]
            day_hour_edge_index = f["day_hour_edge_index"][:]

            YEAR_DAY_EDGE_INDEX, DAY_HOUR_EDGE_INDEX = np.meshgrid(
                year_day_edge_index,
                day_hour_edge_index,
                indexing="ij",
            )

            for attr in f["data"].keys():
                self.region_quantiles[attr] = (
                    f["data"][attr]["normalized_quantiles"][:] * f["data"][attr]["scale"][()] + f["data"][attr]["mean"][()]
                )

            # quantiles["wind_speed"] = np.sqrt(quantiles["wind_east"]**2 + quantiles["wind_north"]**2)

            for attr, data in self.region_quantiles.items():
                y = sp.interpolate.RegularGridInterpolator(
                    (year_day_side, day_hour_side, self.quantile_levels),
                    data[YEAR_DAY_EDGE_INDEX, DAY_HOUR_EDGE_INDEX],
                )((self.utc_year_day, self.utc_day_hour, self.quantiles.get(attr, 0.5)))
                self.data[attr] = y

        wind_speed_correction_factor = self.wind_speed / np.sqrt(
            self.wind_east**2 + self.wind_north**2,
        )
        self.wind_north *= wind_speed_correction_factor
        self.wind_east *= wind_speed_correction_factor

        self.altitude = self.geopotential / g

        self.is_scalar = False

        # handle the overrided parameters
        if "pwv" in self.override.keys():
            force_absolute_humidity = self.override["pwv"] / self.pwv * self.absolute_humidity
            self.humidity = absolute_to_relative_humidity(self.temperature, force_absolute_humidity)

    def __getattr__(self, attr):
        if attr in self.data:
            return self.data[attr]

        raise AttributeError()

    @property
    def absolute_humidity(self):
        return relative_to_absolute_humidity(self.temperature, self.humidity)

    @property
    def dew_point(self):
        return get_dew_point(self.temperature, self.humidity)

    @property
    def wind_bearing(self):
        return np.arctan2(-self.wind_east, self.wind_north) % (2 * np.pi)

    # @property
    # def wind_speed(self):
    #     return np.sqrt(self.wind_east**2 + self.wind_north**2 + self.wind_vertical**2)

    def layers(self):
        df = pd.DataFrame(self.data)
        df.insert(0, "altitude", df.geopotential.values / g)
        df.insert(1, "pressure", 1e2 * self.pressure_levels)

        df = df.loc[df.altitude > self.base_altitude]
        df.index = np.arange(len(df))

        absolute_humidity = relative_to_absolute_humidity(df.temperature, df.humidity)

        df.insert(2, "h_bottom", 0.0)
        df.insert(3, "h_top", 0.0)
        df.insert(4, "h_thickness", 0.0)
        df.insert(5, "total_water", 0.0)
        df.insert(6, "absolute_humidity", absolute_humidity)

        h_values = df.altitude.values
        h_bins = np.array(
            [
                self.base_altitude,
                *(h_values[:-1] + h_values[1:]) / 2,
                h_values[-1] + 100,
            ]
        )

        for layer_index, (h1, h2) in enumerate(zip(h_bins[:-1], h_bins[1:])):
            h = np.linspace(h1, h2, 256)

            absolute_humidity = sp.interpolate.interp1d(
                self.altitude,
                self.absolute_humidity,
                bounds_error=False,
                fill_value="extrapolate",
            )(h)

            df.loc[layer_index, ["h_bottom", "h_top", "h_thickness"]] = h1, h2, h2 - h1
            df.loc[layer_index, "total_water"] = np.trapezoid(absolute_humidity, x=h)

        df.loc[:, "altitude"] = (h_bins[:-1] + h_bins[1:]) / 2

        return df

    @property
    def pwv(self):
        return self.layers().total_water.sum()

    def __call__(self, altitude):
        res = {}
        for field in [*self.fields, "absolute_humidity"]:
            res[field] = sp.interpolate.interp1d(self.altitude, getattr(self, field))(
                altitude,
            )
        return res

    def __repr__(self):
        parts = []
        parts.append(f"region={self.region}")
        parts.append(f"time={self.local_time.format('MMM D HH:mm:ss ZZ')}")
        parts.append(f"altitude={Quantity(self.base_altitude, 'm')}")
        parts.append(f"pwv={Quantity(self.pwv, 'mm')}")
        return f"Weather({', '.join(parts)})"



from __future__ import annotations

import copy

import numpy as np
import pandas as pd

from ..io import DEFAULT_TIME_FORMAT
from ..units import Quantity
from .plan import Plan


class PlanList:
    def __init__(self, plans: list[Plan]):
        if isinstance(plans, PlanList):
            plans = plans.plans
        elif not isinstance(plans, list):
            raise ValueError("'plans' must be a list of Plans")

        self.plans = plans

    def summary(self):
        entries = []
        for p in self.plans:
            c = p.center()
            entry = {
                "start_time": p.start_time.format(DEFAULT_TIME_FORMAT),
                "duration": p.duration,
                # "site": "" if not p.naive else p.site,
                f"target({p.frame.phi_name},{p.frame.theta_name})": c,
            }

            if not p.naive and not p.frame == "az/el":
                entry["center(az,el)"] = str(p.center(frame="az/el"))

            entries.append(entry)

        s = pd.DataFrame(entries)
        s.index.name = "chunk"

        return s

    @property
    def duration(self):
        return Quantity(sum([t.s for t in self.summary().duration]), "s")

    def __repr__(self):
        summary = self.summary()
        return f"""PlanList({len(self.plans)} plans, {self.duration}):
{str(summary)}"""

    def __getitem__(self, index):
        if type(index) is int:
            return self.plans[index]
        elif type(index) is slice:
            return type(self)(self.plans[index])
        else:
            raise ValueError(
                f"Invalid index {index}. A bandList must be indexed by either an integer or a string.",
            )

    def plan_groups(self, max_break: float = 60):
        plan_groups = []
        last_plan_end = -np.inf
        for plan_index, p in enumerate(self.plans):
            if p.start_time.timestamp() - last_plan_end < max_break:
                plan_groups[-1].append(plan_index)
            else:
                plan_groups.append([plan_index])
            last_plan_end = p.end_time.timestamp()

        return plan_groups

    def group_plans(self):
        grouped_plans = []
        for group in self.plan_groups():
            merged_plan = copy.deepcopy(self.plans[group[0]])
            for plan_index in group[1:]:
                merged_plan += self.plans[plan_index]
            grouped_plans.append(merged_plan)

        return PlanList(grouped_plans)



# from __future__ import annotations

# import os

# import numpy as np
# import pandas as pd
# import scipy as sp

# from ..band import BandList
# from ..beam import compute_angular_fwhm
# from ..units import Quantity
# from ..utils import compute_diameter

# here, this_filename = os.path.split(__file__)


# DET_COLUMN_TYPES = {
#     "array": "str",
#     "uid": "str",
#     "array_name": "str",
#     "band_name": "str",
#     "band_center": "float",
#     "sky_x": "float",
#     "sky_y": "float",
#     "baseline_x": "float",
#     "baseline_y": "float",
#     "baseline_z": "float",
#     "polarized": "bool",
#     "pol_angle": "float",
#     "pol_label": "str",
#     "primary_size": "float",
#     "bath_temp": "float",
#     "time_constant": "float",
#     "white_noise": "float",
#     "pink_noise": "float",
#     "efficiency": "float",
# }

# SUPPORTED_ARRAY_PACKINGS = ["hex", "square", "sunflower"]
# SUPPORTED_ARRAY_SHAPES = ["hex", "square", "circle"]


# class ArrayList:
#     def __init__(self, df: pd.DataFrame, bands: BandList, config: dict = {}):
#         self.df = df
#         self.df.index = np.arange(len(self.df.index))

#         self.bands = BandList(bands)
#         self.config = config

#         for band_attr, det_attr in {
#             "center": "band_center",
#             "width": "band_width",
#         }.items():
#             self.df.loc[:, det_attr] = getattr(self, band_attr)

#     def mask(self, **kwargs):
#         mask = np.ones(len(self.df)).astype(bool)
#         for k, v in kwargs.items():
#             mask &= self.df.loc[:, k].values == v
#         return mask

#     def subset(self, **kwargs):
#         return self._subset(self.mask(**kwargs))

#     def _subset(self, mask):
#         df = self.df.loc[mask]
#         return ArrayList(
#             df=df,
#             bands=[b for b in self.bands if b.name in self.df.band_name.values],
#         )

#     def one_detector_from_each_band(self):
#         first_det_mask = np.isin(
#             np.arange(self.n),
#             np.unique(self.band_name, return_index=True)[1],
#         )
#         return self._subset(mask=first_det_mask)

#     def outer(self):
#         outer_dets_index = sp.spatial.ConvexHull(self.offsets).vertices
#         outer_dets_mask = np.isin(np.arange(self.n), outer_dets_index)
#         return self._subset(mask=outer_dets_mask)

#     @property
#     def n(self):
#         return len(self.df)

#     @property
#     def offsets(self):
#         return np.c_[self.sky_x, self.sky_y]

#     @property
#     def baselines(self):
#         return np.c_[self.baseline_x, self.baseline_y, self.baseline_z]

#     @property
#     def field_of_view(self):
#         return Quantity(compute_diameter(self.offsets), "rad")

#     @property
#     def max_baseline(self):
#         return Quantity(compute_diameter(self.baselines), "m")

#     @property
#     def index(self):
#         return self.df.index.values

#     @property
#     def ubands(self):
#         return list(self.bands.keys())

#     @property
#     def fwhm(self):
#         """
#         Returns the angular FWHM (in radians) at infinite distance.
#         """
#         return self.angular_fwhm(z=np.inf)

#     def angular_fwhm(self, z):  # noqa F401
#         """
#         Angular beam width (in radians) as a function of depth (in meters)
#         """
#         return compute_angular_fwhm(z=z, fwhm_0=self.primary_size, n=1, nu=self.band_center.Hz)

#     def physical_fwhm(self, z):
#         """
#         Physical beam width (in meters) as a function of depth (in meters)
#         """
#         return z * self.angular_fwhm(z)

#     def passband(self, nu):
#         _nu = np.atleast_1d(nu)
#         PB = np.zeros((len(self.df), len(_nu)))
#         for band in self.bands:
#             PB[self.band_name == band.name] = band.passband(_nu)
#         return PB

#     def __getattr__(self, attr):
#         if attr in self.df.columns:
#             return self.df.loc[:, attr].values.astype(DET_COLUMN_TYPES[attr])

#         if all(hasattr(band, attr) for band in self.bands):
#             values = np.zeros(shape=self.n, dtype=float)
#             for band in self.bands:
#                 values[self.band_name == band.name] = getattr(band, attr)
#             return values

#         raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

#     def __len__(self):
#         return len(self.df)

#     def __repr__(self):
#         return self.df.__repr__()

#     def _repr_html_(self):
#         return self.df._repr_html_()



from __future__ import annotations

import os

import astropy.constants as const
import numpy as np
from astropy import units as u

from ..constants import T_CMB, c, g, k_B  # noqa
from .angle import Angle  # noqa
from .prefixes import PREFIXES  # noqa
from .quantity import DIMENSIONS, UNITS, Quantity, parse_units  # noqa

here, this_filename = os.path.split(__file__)
symbols = {"radians": "rad"}


def getJynorm():
    factor = 2e26
    factor *= (const.k_B * T_CMB * u.Kelvin) ** 3  # (kboltz*T_CMB)**3.0
    factor /= (const.h * const.c) ** 2  # (hplanck*clight)**2.0
    return factor.value


def getx(freq):
    factor = const.h * freq * u.Hz / const.k_B / (T_CMB * u.Kelvin)
    return factor.to(u.dimensionless_unscaled).value


def KcmbToJy(freq):
    x = getx(freq)
    factor = getJynorm() / T_CMB
    factor *= (x**4) * np.exp(x) / (np.expm1(x) ** 2)
    return factor


# Kelvin CMB to Kelvin brightness
# ----------------------------------------------------------------------
def KcmbToKbright(freq):
    x = getx(freq)
    return np.exp(x) * ((x / np.expm1(x)) ** 2)


# Kelvin brightness to Jy/pixel
# ----------------------------------------------------------------------
def KbrightToJyPix(freq, ipix, jpix=None):
    jpix = jpix or ipix
    return KcmbToJyPix(freq, ipix, jpix) / KcmbToKbright(freq)


# Kelvin CMB to Jy/pixel
# ----------------------------------------------------------------------
def KcmbToJyPix(freq, ipix, jpix=None):
    jpix = jpix or ipix
    x = getx(freq)
    factor = getJynorm() / T_CMB
    factor *= (x**4) * np.exp(x) / (np.expm1(x) ** 2)
    factor *= np.abs(ipix * jpix) * (np.pi / 1.8e2) * (np.pi / 1.8e2)
    return factor



from __future__ import annotations

import glob
import os

import pandas as pd

from ..utils import flatten_config, read_yaml
from .band import Band, get_band, parse_band  # noqa
from .band_list import BandList  # noqa

here, this_filename = os.path.split(__file__)

BAND_FIELD_FORMATS = pd.read_csv(f"{here}/format.csv", index_col=0)

BAND_CONFIGS = {}
for path in glob.glob(f"{here}/configs/*.yml"):
    tag = os.path.split(path)[1].split(".")[0]
    BAND_CONFIGS[tag] = read_yaml(path)

BAND_CONFIGS = flatten_config(BAND_CONFIGS)

band_data = pd.DataFrame(BAND_CONFIGS).T.sort_index()
all_bands = list(band_data.index)



from __future__ import annotations

import matplotlib.pyplot as plt
import pytest

import maria

plt.close("all")


arrays1 = {
    "subarray-1": {
        "n": 500,
        "primary_size": 10,
        "field_of_view": 2,
        "shape": "hexagon",
        "bands": [{"center": 30e9, "width": 5e9}, {"center": 40e9, "width": 5e9}],
    },
    "subarray-2": {
        "n": 500,
        "primary_size": 10,
        "field_of_view": 2,
        "shape": "hexagon",
        "bands": [{"center": 90e9, "width": 5e9}, {"center": 150e9, "width": 5e9}],
        "rotation": 90,
    },
}

# /Users/tom/maria/src/maria/array/

arrays2 = [{"file": "alma/alma_f144.csv", "primary_size": 12, "bands": ["alma/f144"]}]

arrays3 = [{"file": "alma/alma_f144.csv", "primary_size": 12, "bands": ["alma/f144"]}]


array_configs_to_test = [arrays1, arrays2, arrays3]


@pytest.mark.parametrize("array_config", array_configs_to_test)
def test_generate_array(array_config):
    instrument = maria.Instrument(arrays=array_config)



import os

import pandas as pd
import yaml

from ..io import leftpad
from ..spectrum import AtmosphericSpectrum
from ..units import DIMENSIONS, Quantity, parse_units
from .conversion import (
    brightness_temperature_to_cmb_temperature_anisotropy,
    brightness_temperature_to_radiant_flux,
    cmb_temperature_anisotropy_to_brightness_temperature,
    cmb_temperature_anisotropy_to_radiant_flux,
    # cmb_temperature_anisotropy_to_rayleigh_jeans_temperature,
    identity,
    radiant_flux_to_brightness_temperature,
    radiant_flux_to_cmb_temperature_anisotropy,
    radiant_flux_to_rayleigh_jeans_temperature,
    # rayleigh_jeans_temperature_to_cmb_temperature_anisotropy,
    rayleigh_jeans_temperature_to_radiant_flux,
    rayleigh_jeans_temperature_to_spectral_flux_density_per_pixel,
    spectral_flux_density_per_beam_to_spectral_flux_density_per_pixel,
    spectral_flux_density_per_pixel_to_rayleigh_jeans_temperature,
    spectral_flux_density_per_pixel_to_spectral_flux_density_per_beam,
    spectral_flux_density_per_pixel_to_spectral_radiance,
    spectral_radiance_to_spectral_flux_density_per_pixel,
)

here, this_filename = os.path.split(__file__)

conversions = {}
conversions["brightness_temperature"] = {
    "radiant_flux": {"f": brightness_temperature_to_radiant_flux, "linear": False},
    "cmb_temperature_anisotropy": {"f": brightness_temperature_to_cmb_temperature_anisotropy, "linear": False},
}

conversions["radiant_flux"] = {
    "rayleigh_jeans_temperature": {"f": radiant_flux_to_rayleigh_jeans_temperature, "linear": True},
    "cmb_temperature_anisotropy": {"f": radiant_flux_to_cmb_temperature_anisotropy, "linear": True},
    "brightness_temperature": {"f": radiant_flux_to_brightness_temperature, "linear": False},
}
conversions["rayleigh_jeans_temperature"] = {
    "radiant_flux": {"f": rayleigh_jeans_temperature_to_radiant_flux, "linear": True},
    # "cmb_temperature_anisotropy": {"f": rayleigh_jeans_temperature_to_cmb_temperature_anisotropy, "linear": False},
    "spectral_flux_density_per_pixel": {"f": rayleigh_jeans_temperature_to_spectral_flux_density_per_pixel, "linear": False},
}

conversions["cmb_temperature_anisotropy"] = {
    "radiant_flux": {"f": cmb_temperature_anisotropy_to_radiant_flux, "linear": True},
    "brightness_temperature": {"f": cmb_temperature_anisotropy_to_brightness_temperature, "linear": False},
    # "rayleigh_jeans_temperature": {"f": cmb_temperature_anisotropy_to_rayleigh_jeans_temperature, "linear": False},
}

conversions["spectral_flux_density_per_pixel"] = {
    "rayleigh_jeans_temperature": {"f": spectral_flux_density_per_pixel_to_rayleigh_jeans_temperature, "linear": False},
    "spectral_radiance": {"f": spectral_flux_density_per_pixel_to_spectral_radiance, "linear": True},
    "spectral_flux_density_per_beam": {
        "f": spectral_flux_density_per_pixel_to_spectral_flux_density_per_beam,
        "linear": True,
    },
}

conversions["spectral_flux_density_per_beam"] = {
    "spectral_flux_density_per_pixel": {
        "f": spectral_flux_density_per_beam_to_spectral_flux_density_per_pixel,
        "linear": True,
    },
}

conversions["spectral_radiance"] = {
    "spectral_flux_density_per_pixel": {"f": spectral_radiance_to_spectral_flux_density_per_pixel, "linear": True},
}

quantities = list(conversions.keys())
for q in quantities:
    conversions[q][q] = {"f": identity, "linear": True}


walks_dict = {}
paths_dict = {}
for q1 in quantities:
    walks_dict[q1] = [[q1]]
    paths_dict[q1] = {q: [] for q in quantities}

max_steps = 10

steps = 0
while any([paths_dict[q1][q2] == [] for q2 in quantities for q1 in quantities]) and steps < max_steps:
    for start_node in quantities:
        extended_walks = []
        for walk in walks_dict[start_node]:
            for end_node in quantities:
                if conversions.get(walk[-1], {}).get(end_node):
                    extended_walk = [*walk, end_node]
                    extended_walks.append(extended_walk)
                    if not paths_dict[start_node][end_node]:
                        function_path = [conversions[_q2][_q1] for _q1, _q2 in zip(extended_walk[:-1], extended_walk[1:])]
                        # paths_dict[start_node][end_node] = function_path[::-1]
                        paths_dict[start_node][end_node] = extended_walk[::-1]
        walks_dict[start_node] = extended_walks
    steps += 1

calibration_chains = pd.DataFrame(paths_dict)


def parse_calibration_signature(s: str):
    for sep in ["->"]:
        if s.count(sep) == 1:
            if sep is not None:
                items = [u.strip() for u in s.split(sep)]
                if len(items) == 2:
                    res = {}
                    for io, u in zip(["in", "out"], items):
                        res[io] = parse_units(u)
        return res
    raise ValueError("Calibration must have signature 'units1 -> units2'.")


KWARGS_UNITS = {
    "nu": "Hz",
    "pixel_area": "sr",
    "beam_area": "sr",
    "zenith_pwv": "mm",
    "base_temperature": "K",
    "elevation": "rad",
}


class Calibration:
    def __init__(self, signature: str, spectrum: AtmosphericSpectrum = None, **kwargs):
        if not isinstance(signature, str):
            raise ValueError("'signature' must be a string.")

        self.config = pd.DataFrame(parse_calibration_signature(signature))
        self.signature = signature
        self.kwargs = {"spectrum": spectrum, **kwargs}

        for key in kwargs:
            if key not in [
                "nu",
                "polarized",
                "pixel_area",
                "beam_area",
                "band",
                "spectrum",
                "zenith_pwv",
                "base_temperature",
                "elevation",
            ]:
                raise ValueError(f"Invalid kwarg '{key}'.")

    def qchain(self):
        try:
            return calibration_chains.loc[self.in_quantity, self.out_quantity]
        except KeyError:
            raise ValueError(f"Cannot convert from {self.in_quantity} to {self.out_quantity}")

    def uchain(self):
        # middle_terms = [QUANTITIES[q]["default_unit"] for q in self.qchain()[1:-1]]
        middle_terms = [q for q in self.qchain()[1:-1]]
        return " -> ".join([self.in_units, *middle_terms, self.out_units])

    def function_chain(self):
        qchain = self.qchain()
        return [conversions[q1][q2]["f"] for q1, q2 in zip(qchain[:-1], qchain[1:])]

    def linear(self):
        qchain = self.qchain()
        return all([conversions[q2][q1]["linear"] for q1, q2 in zip(qchain[:-1], qchain[1:])])

    def __call__(self, x) -> float:
        y = Quantity(x, self.in_units).base_units_value

        # if self.config.loc["quantity", "in"] == self.config.loc["quantity", "out"]:
        #     return x * self.in_factor / self.out_factor

        # y = x * self.in_factor

        for f in self.function_chain():
            y = f(y, **self.kwargs)

        return Quantity(y, DIMENSIONS.loc[self.qchain()[-1]]).to(self.out_units)

    @property
    def in_units(self) -> float:
        return self.config.loc["units", "in"]

    @property
    def out_units(self) -> float:
        return self.config.loc["units", "out"]

    @property
    def in_factor(self) -> float:
        return self.config.loc["factor", "in"]

    @property
    def out_factor(self) -> float:
        return self.config.loc["factor", "out"]

    @property
    def in_quantity(self) -> float:
        return self.config.loc["quantity", "in"]

    @property
    def out_quantity(self) -> float:
        return self.config.loc["quantity", "out"]

    @property
    def in_to_K_RJ(self) -> float:
        return self.config.loc["from", "in"]

    @property
    def K_RJ_to_out(self) -> float:
        return self.config.loc["to", "out"]

    def leftpad(thing, n: int = 2, char=" "):
        return "\n".join([n * char + line for line in str(thing).splitlines()])

    def __repr__(self):
        KWARGS_UNITS = {"nu": "Hz", "pixel_area": "sr", "zenith_pwv": "mm", "base_temperature": "K", "elevation": "rad"}

        qkwargs = {}
        for k, v in self.kwargs.items():
            if k in ["spectrum", "band"]:
                continue
            qkwargs[k] = str(Quantity(v, KWARGS_UNITS[k]))

        # spectrum_string = self.kwargs.get("spectrum") or ""
        # spectrum_string = self.kwargs.get("spectrum") or ""
        # kwargs_string = yaml.dump(qkwargs)

        factor = f"{self(1e0):.03e}" if self.linear() else "None (nonlinear)"

        return f"""Calibration({self.signature}):
  factor: {factor}
  chain: {self.uchain()}
  spectrum: {self.kwargs.get("spectrum")}
  band: {self.kwargs.get("band")}
  kwargs: {qkwargs}"""

    # def __repr__(self):
    #     filling = copy.deepcopy(self.kwargs.items())
    #     fill_string = ", ".join([self.signature, *[f"{k}={v}" for k, v in self.kwargs.items()]])
    #     return f"Calibration({stuffing})"



from __future__ import annotations

import logging
import os
import time as ttime
from collections.abc import Sequence

import numpy as np
from tqdm import tqdm

from ..io import humanize_time
from ..tod import TOD
from .base import BaseProjectionMapper

np.seterr(invalid="ignore")

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")


class BinMapper(BaseProjectionMapper):
    def __init__(
        self,
        center: tuple[float, float] = None,
        stokes: str = "I",
        width: float = 1,
        height: float = None,
        resolution: float = 0.01,
        frame: str = "ra/dec",
        units: str = "K_RJ",
        degrees: bool = True,
        calibrate: bool = False,
        tod_preprocessing: dict = {},
        map_postprocessing: dict = {},
        tods: Sequence[TOD] = [],
    ):
        height = height or width

        super().__init__(
            center=center,
            stokes=stokes,
            width=width,
            height=height,
            resolution=resolution,
            frame=frame,
            degrees=degrees,
            calibrate=calibrate,
            tods=tods,
            units=units,
            tod_preprocessing=tod_preprocessing,
            map_postprocessing=map_postprocessing,
        )

    def initialize_mapper(self):
        return

    def _run(self):
        """
        The actual mapper for the BinMapper.
        """

        self.map_data = {}

        map_sum = np.zeros((self.n_stokes, len(self.bands), (self.n_y + 2) * (self.n_x + 2)))
        map_wgt = np.zeros((self.n_stokes, len(self.bands), (self.n_y + 2) * (self.n_x + 2)))
        nu_list = []

        for band_index, band in enumerate(self.bands):
            band_start_s = ttime.monotonic()

            nu_list.append(band.center)

            pbar = tqdm(
                enumerate(self.tods),
                total=len(self.tods),
                desc=f"Mapping band {band.name}",
                postfix={"tod": f"1/{len(self.tods)}", "stokes": "I"},
            )

            for tod_index, tod in pbar:
                band_tod = tod.subset(band=band.name)

                if not tod.shape[0] > 0:
                    continue

                band_tod = band_tod.process(config=self.tod_preprocessing).to(self.units)

                P = self.map.pointing_matrix(coords=band_tod.coords)
                D = band_tod.signal.compute()
                W = band_tod.weight.compute()

                stokes_weight = band_tod.dets.stokes_weight()

                for stokes_index, stokes in enumerate(self.stokes):
                    pbar.set_postfix(tod=f"{tod_index + 1}/{len(self.tods)}", stokes=stokes)

                    s = stokes_weight[:, "IQUV".index(stokes)][..., None]

                    map_sum[stokes_index, band_index] += P @ (np.sign(s) * W * D).ravel()
                    map_wgt[stokes_index, band_index] += P @ (np.abs(s) * W).ravel()

                del band_tod

            logger.info(f"Ran mapper for band {band.name} in {humanize_time(ttime.monotonic() - band_start_s)}.")

        return {
            "sum": map_sum.reshape(self.n_stokes, self.n_bands, self.n_y + 2, self.n_x + 2)[..., 1:-1, 1:-1],
            "wgt": map_wgt.reshape(self.n_stokes, self.n_bands, self.n_y + 2, self.n_x + 2)[..., 1:-1, 1:-1],
            "stokes": self.stokes,
            "nu": nu_list,
        }



from __future__ import annotations

import maria
from maria import Planner
from maria.errors import PointingError
from maria.io import fetch


def test_planner():
    input_map = maria.map.load(fetch("maps/crab_nebula.fits"), nu=93e9)

    planner = Planner(
        target=input_map, site="green_bank", constraints={"el": (70, 90), "min_sun_distance": 20, "hour": (14, 15)}
    )

    plans = planner.generate_plans(
        total_duration=900, scan_pattern="daisy", scan_options={"radius": input_map.width.deg / 3}, sample_rate=50
    )

    plans[0].plot()
    print(plans)


def test_planner_error():
    input_map = maria.map.load(fetch("maps/crab_nebula.fits"), nu=93e9)

    planner = Planner(target=input_map, site="amundsen_scott", constraints={"el": (70, 90)})

    try:
        planner.generate_plans(
            total_duration=900,
        )
    except PointingError:
        pass

    else:
        assert False



# AVE MARIA, gratia plena, Dominus tecum.
# Benedicta tu in mulieribus, et benedictus fructus ventris tui, Iesus.
# Sancta Maria, Mater Dei, ora pro nobis peccatoribus, nunc, et in hora mortis nostrae.
from __future__ import annotations

import logging

import pandas

from ._version import __version__, __version_tuple__  # noqa
from .array import Array, all_arrays, get_array  # noqa
from .band import Band, all_bands, get_band  # noqa
from .instrument import Instrument, all_instruments, get_instrument  # noqa
from .io import fetch, set_cache_dir  # noqa
from .map import all_maps  # noqa
from .plan import Plan, Planner, all_plans, get_plan  # noqa
from .sim import Simulation  # noqa
from .site import Site, all_regions, all_sites, get_site, site_data  # noqa
from .units import Quantity  # noqa

logging.getLogger("jax._src.xla_bridge").setLevel(logging.ERROR)

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s.%(msecs)03d %(levelname)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

logger = logging.getLogger("maria")

pandas.set_option("display.expand_frame_repr", False)


def debug():
    logger.setLevel(logging.DEBUG)


def undebug():
    logger.setLevel(logging.INFO)



from __future__ import annotations

import numpy as np
import pytest

import maria
from maria import Simulation
from maria.atmosphere import Atmosphere, AtmosphericSpectrum


@pytest.mark.parametrize("region_name", maria.all_regions)
def test_atmosphere(region_name):
    atmosphere = Atmosphere(region=region_name)
    atmosphere.spectrum.emission(nu=90e9, elevation=1.1, pwv=1.5)


@pytest.mark.parametrize("region_name", ["chajnantor"])
def test_spectrum_from_cache(region_name):
    spectrum = AtmosphericSpectrum(region=region_name, refresh_cache=True)  # noqa


def test_atmosphere_2d():
    sim = Simulation(
        instrument="MUSTANG-2",
        plans="ten_second_zenith_stare",
        site="green_bank",
        atmosphere="2d",
    )
    tods = sim.run()  # noqa



from __future__ import annotations

import os

import numpy as np
import pytest

import maria
from maria import Simulation, all_instruments, all_sites, get_plan
from maria.mappers import BinMapper
from maria.utils import read_yaml

here, this_filename = os.path.split(__file__)

all_test_plan_configs = list(read_yaml(f"{here}/configs/test_plans.yml").values())

all_instruments.pop(all_instruments.index("alma/ALMA"))

n_sims = 3
test_instruments = np.random.choice(a=all_instruments, size=n_sims)
test_sites = np.random.choice(a=all_sites, size=n_sims)


@pytest.mark.parametrize(
    "instrument,site",
    zip(test_instruments, test_sites),
)
def test_pipeline(instrument, site):
    site = maria.get_site(site)

    plan = maria.Plan.generate(
        start_time="2024-08-06T09:00:00",
        scan_pattern="daisy",
        scan_options={"radius": 0.5, "speed": 0.1},  # in degrees
        duration=5,  # in seconds
        sample_rate=20,  # in Hz
        scan_center=(31, 62),
        frame="az/el",
    )

    sim = Simulation(
        instrument=instrument,
        site=site,
        plans=[plan],
        atmosphere="2d",
        cmb="generate",
    )

    tod = sim.run()[0]

    for field in ["atmosphere", "cmb"]:
        if np.isnan(tod.data[field]).any():
            raise ValueError(f"There are NaNs in the '{field}' field.")

    tod = tod.to("K_RJ")

    mapper = BinMapper(
        center=(0, -23),
        stokes="IQUV",
        frame="ra/dec",
        width=1,
        height=1,
        resolution=1 / 256,
        tod_preprocessing={
            "window": {"name": "tukey", "kwargs": {"alpha": 0.1}},
            "remove_spline": {"knot_spacing": 30, "remove_el_gradient": True},
            "remove_modes": {"modes_to_remove": [0]},
        },
        map_postprocessing={
            "gaussian_filter": {"sigma": 1},
            "median_filter": {"size": 1},
        },
        units="mK_RJ",
    )

    mapper.add_tods(tod)
    mapper.run()



import logging
import time as ttime

import numpy as np

from ..units import Quantity


def humanize(x, units):
    return str(Quantity(x, units=units))


def humanize_time(seconds):
    return humanize(seconds, units="s")


def log_duration(ref_time, message, level="debug"):
    logger = logging.getLogger("maria")
    string = f"{message} in {humanize_time(ttime.monotonic() - ref_time)}."
    getattr(logger, level)(string)


def leftpad(thing, n: int = 2, char=" "):
    return "\n".join([n * char + line for line in str(thing).splitlines()])



from __future__ import annotations

import os

import numpy as np
import pytest

import maria
from maria import Planner, Simulation, all_instruments
from maria.io import fetch
from maria.mappers import BinMapper
from maria.utils import read_yaml

here, this_filename = os.path.split(__file__)


def test_polarized_map_sim():
    einstein = maria.map.load(fetch("maps/einstein.h5"))
    planner = Planner(target=einstein, site="llano_de_chajnantor", constraints={"el": (60, 90)})
    plans = planner.generate_plans(total_duration=10, sample_rate=50)  # in Hz

    plans[0].plot()
    print(plans)

    sim = Simulation(
        instrument="test/1deg",
        site="llano_de_chajnantor",
        plans=plans,
        atmosphere="2d",
        cmb="generate",
        map=einstein,
    )

    tod = sim.run()[0]

    for field in ["atmosphere", "cmb"]:
        if np.isnan(tod.data[field]).any():
            raise ValueError(f"There are NaNs in the '{field}' field.")

    tod = tod.to("K_RJ")

    mapper = BinMapper(
        center=(0, -23),
        stokes="IQUV",
        frame="ra/dec",
        width=1,
        height=1,
        resolution=1 / 256,
        tod_preprocessing={
            "window": {"name": "tukey", "kwargs": {"alpha": 0.1}},
            "remove_spline": {"knot_spacing": 30, "remove_el_gradient": True},
            "remove_modes": {"modes_to_remove": [0]},
        },
        map_postprocessing={
            "gaussian_filter": {"sigma": 1},
            "median_filter": {"size": 1},
        },
        units="mK_RJ",
    )

    mapper.add_tods(tod)
    mapper.run()



import numpy as np
import scipy as sp

from .radiometry import *  # noqa


def sigmoid(x):
    return 1 / (1 + np.exp(-x))


def inverse_sigmoid(y):
    return -np.log(1 / y - 1)


def matern(r, r0, nu):
    """
    Matérn covariance
    """
    return normalized_matern(r / r0, nu)


def normalized_matern(r, nu):
    """
    Normalized Matérn covariance
    """
    return (
        2 ** (1 - nu)
        / sp.special.gamma(nu)
        * sp.special.kv(nu, np.sqrt(2 * nu) * r + 1e-16)
        * (np.sqrt(2 * nu) * r + 1e-16) ** nu
    )


def approximate_normalized_matern(r, nu=1 / 3, r0=1e0, n_test_points=1024):
    """
    Computing BesselK[nu,z] for arbitrary nu is expensive. This is good for casting over huge matrices.
    """
    r_eff = r / r0
    r_eff_min, r_eff_max = 1e-6, 1e3

    r_eff_safe = np.atleast_1d(np.abs(r_eff)).clip(min=r_eff_min)
    r_eff_nonzero = r_eff_safe[r_eff_safe < r_eff_max]

    r_eff_samples = np.geomspace(r_eff_min, r_eff_max, n_test_points)
    cov_samples = normalized_matern(r_eff_samples, nu=nu)

    with np.errstate(divide="ignore"):
        sf = np.exp(
            np.interp(
                np.log(r_eff_nonzero),
                np.log(r_eff_samples),
                np.log(1 - cov_samples),
            ),
        )
        cov = np.exp(
            np.interp(np.log(r_eff_nonzero), np.log(r_eff_samples), np.log(cov_samples)),
        )

    # we combine the log interpolations so that both extremes have really good precision
    # the structure function and covariance are equal at around 1, so that's our inflection point
    t = 1 / (1 + r_eff_nonzero**2)

    res = np.zeros(r.shape)
    res[r_eff_safe < r_eff_max] = t * (1 - sf) + (1 - t) * cov

    return res



from __future__ import annotations

import os
from pathlib import Path

import h5py
import healpy as hp
import matplotlib as mpl
import numpy as np
import pandas as pd
from astropy.coordinates import EarthLocation
from matplotlib import pyplot as plt

from ..constants import earth_radius
from ..io import fetch, repr_lat_lon
from ..units import Quantity

here, this_filename = os.path.split(__file__)

REGION_DISPLAY_COLUMNS = ["location", "country", "latitude", "longitude", "timezone"]
REGIONS = pd.read_csv(f"{here}/regions.csv", index_col=0)
all_regions = list(REGIONS.index.values)


class InvalidRegionError(Exception):
    def __init__(self, invalid_region):
        super().__init__(
            f"The region '{invalid_region}' is not supported. "
            f"Supported regions are:\n\n{REGIONS.loc[:, REGION_DISPLAY_COLUMNS].to_string()}",
        )


def get_height_map():
    with h5py.File(fetch("world_heightmap.h5"), "r") as f:
        height_map = f["data"][:].astype(np.uint16)
    return 32 * np.where(height_map < 255, height_map, np.nan)


class Site:
    """
    An object representing a point on the earth that parametrizes observing conditions.
    """

    def __init__(
        self,
        description: str = "",
        region: str = "princeton",
        timezone: str = "UTC",
        seasonal: bool = True,
        diurnal: bool = True,
        latitude: float = None,  # in degrees
        longitude: float = None,  # in degrees
        altitude: float = None,  # in meters
        weather_quantiles: dict = {},
        documentation: str = "",
        instruments: list = [],
    ):
        if region not in REGIONS.index.values:
            raise InvalidRegionError(region)

        if longitude is None:
            longitude = Quantity(REGIONS.loc[region].longitude, "deg")

        if latitude is None:
            latitude = Quantity(REGIONS.loc[region].latitude, "deg")

        if altitude is None:
            altitude = Quantity(REGIONS.loc[region].altitude, "m")

        self.description = description
        self.region = region
        self.timezone = REGIONS.loc[region, "timezone"] if region in all_regions else timezone
        self.seasonal = seasonal
        self.diurnal = diurnal
        self.latitude = Quantity(latitude, "deg")
        self.longitude = Quantity(longitude, "deg")
        self.altitude = Quantity(altitude, "m")
        self.weather_quantiles = weather_quantiles
        self.documentation = documentation
        self.instruments = instruments

        self.earth_location = EarthLocation.from_geodetic(
            lon=self.longitude.deg,
            lat=self.latitude.deg,
            height=self.altitude.m,
        )

    def plot(self, res=0.03, wide_size: float = 25, zoom_size: float = 2, cmap="cubehelix"):
        height_map = get_height_map()

        kwargs = {
            "rot": (self.longitude.deg, self.latitude.deg),
            "cmap": cmap,
            "flip": "geo",
            "badcolor": "gray",
            "reso": 60 * res,
            "return_projected_map": True,
            "no_plot": True,
        }

        wide_map = hp.gnomview(
            height_map,
            xsize=wide_size / res,
            **kwargs,
        )[::2, ::2]

        zoom_map = hp.gnomview(
            height_map,
            xsize=zoom_size / res,
            **kwargs,
        )

        fig, axes = plt.subplots(1, 2, figsize=(8, 5), constrained_layout=True)

        wide_vmax = np.nanpercentile(wide_map.data, q=99.9)
        zoom_vmax = np.nanpercentile(zoom_map.data, q=95)

        cmap = plt.get_cmap(cmap)
        cmap.set_bad("royalblue", 1.0)

        wide_mesh = axes[0].pcolormesh(
            1e-3 * earth_radius * np.radians(wide_size) * np.linspace(-0.5, 0.5, wide_map.shape[0]),
            1e-3 * earth_radius * np.radians(wide_size) * np.linspace(-0.5, 0.5, wide_map.shape[1]),
            wide_map,
            cmap=cmap,
            vmax=wide_vmax,
            shading="nearest",
        )
        cbar = fig.colorbar(wide_mesh, ax=axes[0], location="bottom", shrink=0.9, label="height")
        cbar.set_label("Height [meters]")

        zoom_mesh = axes[1].pcolormesh(
            1e-3 * earth_radius * np.radians(zoom_size) * np.linspace(-0.5, 0.5, zoom_map.shape[0]),
            1e-3 * earth_radius * np.radians(zoom_size) * np.linspace(-0.5, 0.5, zoom_map.shape[1]),
            zoom_map,
            cmap=cmap,
            vmax=zoom_vmax,
            shading="nearest",
        )
        cbar = fig.colorbar(zoom_mesh, ax=axes[1], location="bottom", shrink=0.9, label="height")
        cbar.set_label("Height [meters]")

        for ax in axes:
            ax.scatter(0, 0, edgecolor="r", facecolor="none", linewidth=2, s=256)
            ax.set_xlabel("Distance [kilometers]")
            ax.set_yticks([])
            ax.set_facecolor("gray")
            ax.set_rasterized(True)
            ax.set_aspect("equal")

    @property
    def location(self):
        return (self.longitude, self.latitude, self.altitude)

    def __repr__(self):
        repr_lat, repr_lon = repr_lat_lon(self.latitude.degrees, self.longitude.degrees)
        s = f"""Site:
  region: {self.region}
  timezone: {self.timezone}
  location:
    longitude: {repr_lon}
    latitude:  {repr_lat}
    altitude: {self.altitude}
  seasonal: {self.seasonal}
  diurnal: {self.diurnal}"""

        return s
